import logging
from typing import Dict, List

from moodle_dl.config import ConfigHelper
from moodle_dl.moodle.mods import MoodleMod
from moodle_dl.types import Course, File


class DataMod(MoodleMod):
    MOD_NAME = 'data'
    MOD_PLURAL_NAME = 'databases'
    MOD_MIN_VERSION = 2015051100  # 2.9

    @classmethod
    def download_condition(cls, config: ConfigHelper, file: File) -> bool:
        return config.get_download_databases() or file.content_type != 'database_file'

    async def real_fetch_mod_entries(
        self, courses: List[Course], core_contents: Dict[int, List[Dict]]
    ) -> Dict[int, Dict[int, Dict]]:
        databases = (
            await self.client.async_post(
                'mod_data_get_databases_by_courses', self.get_data_for_mod_entries_endpoint(courses)
            )
        ).get('databases', [])

        result = {}
        for database in databases:
            course_id = database.get('course', 0)
            database_files = database.get('introfiles', [])
            self.set_props_of_files(database_files, type='database_introfile')

            database_intro = database.get('intro', '')
            if database_intro != '':
                database_files.append(
                    {
                        'filename': 'Database intro',
                        'filepath': '/',
                        'description': database_intro,
                        'type': 'description',
                    }
                )

            self.add_module(
                result,
                course_id,
                database.get('coursemodule', 0),
                {
                    'id': database.get('id', 0),
                    'name': database.get('name', 'db'),
                    'files': database_files,
                },
            )

        await self.add_database_files(result)
        return result

    async def add_database_files(self, databases: Dict[int, Dict[int, Dict]]):
        """
        Fetches for the databases list the database file entries.
        @param databases: Dictionary of all databases, indexed by courses, then module id
        """
        if not self.config.get_download_databases():
            return

        if self.version < 2017051500:  # 3.3
            return

        await self.run_async_load_function_on_mod_entries(databases, self.load_database_files)

    async def load_database_files(self, database: Dict):
        "Fetches for a given assign database the database files"
        data = {'databaseid': database.get('id', 0)}
        access = await self.client.async_post('mod_data_get_data_access_information', data)
        if not access.get('timeavailable', False):
            logging.debug("No access rights for database %d", database.get('id', 0))
            return

        data.update({'returncontents': 1})
        entries = await self.client.async_post('mod_data_get_entries', data)
        database['files'] += self._get_files_of_db_entries(entries)

    @classmethod
    def _get_files_of_db_entries(cls, entries: Dict) -> List:
        result = []
        for entry in entries.get('entries', []):
            for entry_content in entry.get('contents', []):
                for entry_file in entry_content.get('files', []):
                    filename = entry_file.get('filename', '')
                    if filename.startswith('thumb_'):
                        continue
                    cls.set_props_of_file(entry_file, type='database_file')
                    result.append(entry_file)

        return result



import logging
from typing import Dict, List

from moodle_dl.config import ConfigHelper
from moodle_dl.moodle.mods import MoodleMod
from moodle_dl.moodle.moodle_constants import moodle_html_footer, moodle_html_header
from moodle_dl.moodle.request_helper import RequestRejectedError
from moodle_dl.types import Course, File
from moodle_dl.utils import PathTools as PT


class QuizMod(MoodleMod):
    MOD_NAME = 'quiz'
    MOD_PLURAL_NAME = 'quizzes'
    MOD_MIN_VERSION = 2016052300  # 3.1

    @classmethod
    def download_condition(cls, config: ConfigHelper, file: File) -> bool:
        return config.get_download_quizzes() or (not (file.module_modname.endswith(cls.MOD_NAME) and file.deleted))

    async def real_fetch_mod_entries(
        self, courses: List[Course], core_contents: Dict[int, List[Dict]]
    ) -> Dict[int, Dict[int, Dict]]:
        quizzes = (
            await self.client.async_post(
                'mod_quiz_get_quizzes_by_courses', self.get_data_for_mod_entries_endpoint(courses)
            )
        ).get('quizzes', [])

        result = {}
        for quiz in quizzes:
            course_id = quiz.get('course', 0)

            quiz_files = quiz.get('introfiles', [])
            self.set_props_of_files(quiz_files, type='quiz_introfile')

            quiz_intro = quiz.get('intro', '')
            if quiz_intro != '':
                quiz_files.append(
                    {
                        'filename': 'Quiz intro',
                        'filepath': '/',
                        'description': quiz_intro,
                        'type': 'description',
                    }
                )

            self.add_module(
                result,
                course_id,
                quiz.get('coursemodule', 0),
                {
                    'id': quiz.get('id', 0),
                    'name': quiz.get('name', 'unnamed quiz'),
                    'files': quiz_files,
                },
            )

        await self.add_quizzes_files(result)
        return result

    async def add_quizzes_files(self, quizzes: Dict[int, Dict[int, Dict]]):
        """
        Fetches for the quizzes list the quizzes files
        @param quizzes: Dictionary of all quizzes, indexed by courses, then module id
        """
        if not self.config.get_download_quizzes():
            return

        if self.version < 2016052300:  # 3.1
            return

        await self.run_async_load_function_on_mod_entries(quizzes, self.load_quiz_files)

    async def load_quiz_files(self, quiz: Dict):
        data = {'quizid': quiz.get('id', 0), 'userid': self.user_id, 'status': 'all'}
        attempts = (await self.client.async_post('mod_quiz_get_user_attempts', data)).get('attempts', [])
        quiz_name = quiz.get('name', '')
        for attempt in attempts:
            attempt['_quiz_name'] = quiz_name

        quiz['files'] += await self.run_async_collect_function_on_list(
            attempts,
            self.load_files_of_attempt,
            'attempt',
            {'collect_id': 'id', 'collect_name': '_quiz_name'},
        )

    async def load_files_of_attempt(self, attempt: Dict) -> List[Dict]:
        result = []

        attempt_id = attempt.get('id', 0)
        attempt_state = attempt.get('state', 'unknown')
        quiz_name = attempt.get('_quiz_name', '')

        attempt_filename = PT.to_valid_name(
            quiz_name + ' (attempt ' + str(attempt_id) + ' ' + attempt_state + ')', is_file=False
        )

        data = {'attemptid': attempt_id}
        try:
            if attempt_state == 'finished':
                questions = (await self.client.async_post('mod_quiz_get_attempt_review', data)).get('questions', [])
            elif attempt_state == 'inprogress':
                questions = (await self.client.async_post('mod_quiz_get_attempt_summary', data)).get('questions', [])
            else:
                return result
        except RequestRejectedError:
            logging.debug("No access rights for quiz attempt %d", attempt_id)
            return result

        # build quiz HTML
        quiz_html = moodle_html_header
        for question in questions:
            question_html = question.get('html', '').split('<script>')[0]
            if question_html is None:
                question_html = ''
            quiz_html += question_html + '\n'

            question_files = question.get('responsefileareas', [])
            self.set_props_of_files(question_files, type='quiz_file')
            result.extend(question_files)

        quiz_html += moodle_html_footer
        result.append(
            {
                'filename': attempt_filename,
                'filepath': '/',
                'timemodified': attempt.get('timemodified', 0),
                'html': quiz_html,
                'type': 'html',
                'no_search_for_urls': True,
            }
        )

        return result



import json
from urllib.parse import parse_qs, urlencode, urljoin, urlparse, urlunparse

from yt_dlp.extractor.common import InfoExtractor
from yt_dlp.utils import traverse_obj


class SharePointIE(InfoExtractor):
    _VALID_URL = r'https?://[^.]+\.sharepoint\.com/:v:/g/(?:[^/]+/){2}(?P<id>[^/?#]+)'

    _TESTS = [
        {
            'url': 'https://lut-my.sharepoint.com/:v:/g/personal/juha_eerola_student_lab_fi/'
            + 'EUrAmrktb4ZMhUcY9J2PqMEBD_9x_l0DyYWVgAvp-TTOMw?e=ZpQOOw',
            'info_dict': {
                'id': 'EUrAmrktb4ZMhUcY9J2PqMEBD_9x_l0DyYWVgAvp-TTOMw',
                'ext': 'unknown_video',
                'title': 'CmvpJST.mp4',
                'duration': 54.567,
            },
            'params': {'skip_download': 'm3u8'},
        },
    ]

    def _real_extract(self, url):
        video_id = self._match_id(url)
        webpage = self._download_webpage(url, video_id)

        video = self._search_json(r'g_fileInfo\s*=', webpage, 'player config', video_id)

        parsed_url = urlparse(video['.transformUrl'])
        base_media_url = urlunparse(
            parsed_url._replace(
                path=urljoin(f'{parsed_url.path}/', '../videomanifest'),
                query=urlencode(
                    {
                        **parse_qs(parsed_url.query),
                        'cTag': video['.ctag'],
                        'action': 'Access',
                        'part': 'index',
                    },
                    doseq=True,
                ),
            )
        )

        # The player is configured with a number of options that are then added to the format URLs,
        # but we always get all formats back no matter what options we send
        hls_formats = self._extract_m3u8_formats(f'{base_media_url}&format=hls', video_id, m3u8_id='hls', fatal=False)
        hls_formats.extend(
            self._extract_m3u8_formats(f'{base_media_url}&format=hls-vnext', video_id, m3u8_id='hls-vnext', fatal=False)
        )
        for hls_format in hls_formats:
            hls_format['preference'] = -2
        formats = self._extract_mpd_formats(f'{base_media_url}&format=dash', video_id, mpd_id='dash', fatal=False)
        formats.extend(hls_formats)

        return {
            'id': video_id,
            'formats': formats,
            'title': video.get('name'),
            'duration': traverse_obj(
                video, ('MediaServiceFastMetadata', {json.loads}, 'media', 'duration', {lambda x: x / 10000000})
            ),
        }



import os
import sys

from moodle_dl.cli.config_wizard import ConfigWizard
from moodle_dl.cli.database_manager import DatabaseManager
from moodle_dl.cli.moodle_wizard import MoodleWizard
from moodle_dl.cli.notifications_wizard import NotificationsWizard
from moodle_dl.config import ConfigHelper
from moodle_dl.types import MoodleDlOpts
from moodle_dl.utils import Cutie, Log

__all__ = ["ConfigWizard", "DatabaseManager", "NotificationsWizard"]


def init_config(config: ConfigHelper, opts: MoodleDlOpts):
    if config.is_present():
        do_override_input = Cutie.prompt_yes_or_no(Log.error_str('Do you want to override the existing config?'))

        if not do_override_input:
            sys.exit(0)

    NotificationsWizard(config, opts).interactively_configure_all_services()

    do_sentry = Cutie.prompt_yes_or_no('Do you want to configure Error Reporting via Sentry?')
    if do_sentry:
        sentry_dsn = input('Please enter your Sentry DSN:   ')
        config.set_property('sentry_dsn', sentry_dsn)

    MoodleWizard(config, opts).interactively_acquire_token()

    Log.success('Configuration finished and saved!')

    if os.name != 'nt':
        working_dir = os.path.abspath(opts.path)
        moodle_dl_path = os.path.abspath(sys.argv[0])
        Log.info(
            '  To set a cron-job for this program on your Unix-System:\n'
            + '    1. `crontab -e`\n'
            + f'    2. Add `*/15 * * * * cd "{working_dir}" && "{moodle_dl_path}" >/dev/null 2>&1`\n'
            + '    3. Save and you\'re done!'
        )

        Log.info(
            'For more ways to run `moodle-dl` periodically, take a look at the wiki'
            + ' (https://github.com/C0D3D3V/Moodle-DL/wiki/Start-Moodle-dl-periodically-or-via-Telegram)'
        )
    else:
        Log.info(
            'If you want to run moodle-dl periodically, you can take a look at the wiki '
            + '(https://github.com/C0D3D3V/Moodle-DL/wiki/Start-Moodle-dl-periodically-or-via-Telegram)'
        )

    print('')

    Log.info('You can always do the additional configuration later with the --config option.')

    do_config = Cutie.prompt_yes_or_no('Do you want to make additional configurations now?')
    if do_config:
        ConfigWizard(config, opts).interactively_acquire_config()

    print('')
    Log.success('All set and ready to go!')



import base64
import logging
import re
from typing import List, Tuple
from urllib.parse import urlparse

from moodle_dl.config import ConfigHelper
from moodle_dl.database import StateRecorder
from moodle_dl.moodle.cookie_handler import CookieHandler
from moodle_dl.moodle.core_handler import CoreHandler
from moodle_dl.moodle.mods import (
    fetch_mods_files,
    get_all_mods,
    get_all_mods_classes,
    get_mod_plurals,
)
from moodle_dl.moodle.request_helper import RequestHelper
from moodle_dl.moodle.result_builder import ResultBuilder
from moodle_dl.types import Course, MoodleDlOpts, MoodleURL
from moodle_dl.utils import determine_ext


class MoodleService:
    def __init__(self, config: ConfigHelper, opts: MoodleDlOpts):
        self.config = config
        self.opts = opts

    def obtain_login_token(self, username: str, password: str, moodle_url: MoodleURL) -> str:
        "Send the login credentials to the Moodle-System and extracts the resulting Login-Token"
        login_data = {'username': username, 'password': password, 'service': 'moodle_mobile_app'}
        response = RequestHelper(self.config, self.opts, moodle_url, None).get_login(login_data)

        if 'token' not in response:
            # = we didn't get an error page (checked by the RequestHelper) but
            # somehow we don't have the needed token
            raise RuntimeError('Invalid response received from the Moodle System!  No token was received.')

        if 'privatetoken' not in response:
            return response.get('token', ''), None
        return response.get('token', ''), response.get('privatetoken', '')

    @staticmethod
    def extract_token(address: str) -> str:
        """
        Extracts a token from a returned URL
        See https://github.com/moodle/moodle/blob/master/admin/tool/mobile/launch.php for details
        """
        splitted = address.split('token=')

        if len(splitted) < 2:
            return None

        decoded = str(base64.b64decode(splitted[1]))

        splitted = decoded.split(':::')
        if len(splitted) < 2:
            return None

        token = re.sub(r'[^A-Za-z0-9]+', '', splitted[1])

        if len(splitted) < 3:
            return token, None

        secret_token = re.sub(r'[^A-Za-z0-9]+', '', splitted[2])
        return (token, secret_token)

    def get_courses_list(self, core_handler: CoreHandler, user_id: int) -> List[Course]:
        download_course_ids = self.config.get_download_course_ids()
        download_public_course_ids = self.config.get_download_public_course_ids()
        dont_download_course_ids = self.config.get_dont_download_course_ids()

        courses_list = core_handler.fetch_courses(user_id)
        courses = []
        # Filter unselected courses
        for course in courses_list:
            if MoodleService.should_download_course(course.id, download_course_ids, dont_download_course_ids):
                courses.append(course)

        public_courses_list = core_handler.fetch_courses_info(download_public_course_ids)
        courses.extend(public_courses_list)
        return courses

    def get_user_id_and_version(self, core_handler: CoreHandler) -> Tuple[int, int]:
        user_id, version = self.config.get_userid_and_version()
        if user_id is None or version is None:
            logging.info('Downloading account information')
            user_id, version = core_handler.fetch_userid_and_version()
            logging.debug('Detected moodle version: %d', version)
        else:
            core_handler.version = version
        return user_id, version

    async def fetch_state(self, database: StateRecorder) -> List[Course]:
        """
        Fetch the current status of the configured Moodle account and compare it with the last known state
        It does not change the known state, nor does it download the files.
        @return: List with detected changes between the new and old state
        """
        logging.debug('Fetching current Moodle state...')
        token = self.config.get_token()
        privatetoken = self.config.get_privatetoken()
        moodle_url = self.config.get_moodle_URL()

        request_helper = RequestHelper(self.config, self.opts, moodle_url, token)
        core_handler = CoreHandler(request_helper)
        user_id, version = self.get_user_id_and_version(core_handler)

        cookie_handler = None
        if self.config.get_download_also_with_cookie():
            cookie_handler = CookieHandler(request_helper, version, self.config, self.opts)
            cookie_handler.check_and_fetch_cookies(privatetoken, user_id)

        courses = self.get_courses_list(core_handler, user_id)

        core_contents = await core_handler.async_load_core_contents(courses)
        mods = get_all_mods(request_helper, version, user_id, database.get_last_timestamp_per_mod_module(), self.config)
        fetched_mods_files = await fetch_mods_files(mods, courses, core_contents)

        logging.debug('Combine API results...')
        ResultBuilder(moodle_url, version, get_mod_plurals()).add_files_to_courses(
            courses, core_contents, fetched_mods_files
        )

        logging.debug('Checking for changes...')
        changes = database.changes_of_new_version(courses)
        changes = self.add_options_to_courses(changes)
        changes = self.filter_courses(changes, self.config, cookie_handler, courses)

        return changes

    def add_options_to_courses(self, courses: List[Course]):
        "Updates the courses with their options"
        options_of_courses = self.config.get_options_of_courses()
        for course in courses:
            options = options_of_courses.get(str(course.id), None)
            if options is not None:
                course.overwrite_name_with = options.get('overwrite_name_with', None)
                course.create_directory_structure = options.get('create_directory_structure', True)
                course.excluded_sections = options.get("excluded_sections", [])

        return courses

    @staticmethod
    def filter_courses(
        changes: List[Course],
        config: ConfigHelper,
        cookie_handler: CookieHandler = None,
        courses_list: List[Course] = None,
    ) -> List[Course]:
        """
        Filters the changes course list from courses that
        should not get downloaded
        @param config: ConfigHelper to obtain all the different filter configs
        @param cookie_handler: CookieHandler to check if the cookie is valid
        @param courses_list: A list of all courses that are available online
        @return: filtered changes course list
        """

        download_course_ids = config.get_download_course_ids()
        dont_download_course_ids = config.get_dont_download_course_ids()
        download_public_course_ids = config.get_download_public_course_ids()
        download_descriptions = config.get_download_descriptions()
        download_links_in_descriptions = config.get_download_links_in_descriptions()
        exclude_file_extensions = config.get_exclude_file_extensions()
        max_file_size = config.get_max_file_size()

        download_also_with_cookie = config.get_download_also_with_cookie()
        if cookie_handler is not None:
            download_also_with_cookie = cookie_handler.test_cookies()

        all_mods_classes = get_all_mods_classes()
        filtered_changes = []

        for course in changes:
            if not MoodleService.should_download_course(
                course.id, download_course_ids + download_public_course_ids, dont_download_course_ids
            ):
                # Filter courses that should not be downloaded
                continue

            if courses_list is not None:
                not_online = True
                # Filter courses that are not available online
                for online_course in courses_list:
                    if online_course.id == course.id:
                        not_online = False
                        break
                if not_online:
                    logging.warning('The Moodle course with id %d is no longer available online.', course.id)
                    continue

            course_files = []
            for file in course.files:
                # Filter files based on module options
                modules_conditions_met = True
                for mod in all_mods_classes:
                    if not mod.download_condition(config, file):
                        modules_conditions_met = False
                        break

                # Filter Files based on other options
                if (
                    modules_conditions_met
                    # Filter Description Files (except the forum posts)
                    and (
                        download_descriptions
                        or file.content_type != 'description'
                        or (
                            file.module_modname == 'forum'
                            and file.content_type == 'description'
                            and file.content_filename != 'Forum intro'
                        )
                    )
                    # Filter Files that requiere a Cookie
                    and (download_also_with_cookie or (not file.module_modname.startswith('cookie_mod-')))
                    # Exclude files whose file extension is blacklisted
                    and (determine_ext(file.content_filename) not in exclude_file_extensions)
                    # Exclude files that are in excluded sections
                    and (MoodleService.should_download_section(file.section_id, course.excluded_sections))
                    # Exclude files that are bigger than max_file_size
                    and (max_file_size == 0 or file.content_filesize < max_file_size)
                ):
                    course_files.append(file)
            course.files = course_files

            # Filter Description URLs
            course_files = []
            for file in course.files:
                if not file.content_type == 'description-url':
                    course_files.append(file)

                elif download_links_in_descriptions:
                    add_description_url = True
                    for test_file in course.files:
                        if file.content_fileurl == test_file.content_fileurl:
                            if test_file.content_type != 'description-url':
                                # If a URL in a description also exists as a real link in the course,
                                # then ignore this URL
                                add_description_url = False
                                break
                            if file.module_id > test_file.module_id:
                                # Always use the link from the older description.
                                add_description_url = False
                                break

                    if add_description_url:
                        course_files.append(file)
            course.files = course_files

            if len(course.files) > 0:
                filtered_changes.append(course)

        return filtered_changes

    @staticmethod
    def should_download_course(
        course_id: int, download_course_ids: List[int], dont_download_course_ids: List[int]
    ) -> bool:
        "Checks if a course is in whitelist and not in blacklist"
        inBlacklist = course_id in dont_download_course_ids
        inWhitelist = course_id in download_course_ids or len(download_course_ids) == 0

        return inWhitelist and not inBlacklist

    @staticmethod
    def should_download_section(section_id: int, dont_download_sections_ids: List[int]) -> bool:
        "Checks if a section is not in blacklist"
        return section_id not in dont_download_sections_ids or len(dont_download_sections_ids) == 0

    @staticmethod
    def split_moodle_url(moodle_url: str) -> Tuple[str, str]:
        """
        Splits a given Moodle URL into the domain and the installation path
        @return: moodle_domain, moodle_path as strings
        """
        moodle_uri = urlparse(moodle_url)
        moodle_domain = moodle_uri.netloc
        moodle_path = moodle_uri.path
        if not moodle_path.endswith('/'):
            moodle_path = moodle_path + '/'

        if moodle_path == '':
            moodle_path = '/'

        return moodle_domain, moodle_path



import logging
from typing import Dict, List

from moodle_dl.config import ConfigHelper
from moodle_dl.moodle.mods import MoodleMod
from moodle_dl.moodle.request_helper import RequestRejectedError
from moodle_dl.types import Course, File


class WorkshopMod(MoodleMod):
    MOD_NAME = 'workshop'
    MOD_PLURAL_NAME = 'workshops'
    MOD_MIN_VERSION = 2017111300  # 3.4

    @classmethod
    def download_condition(cls, config: ConfigHelper, file: File) -> bool:
        return config.get_download_workshops() or (not (file.module_modname.endswith(cls.MOD_NAME) and file.deleted))

    async def real_fetch_mod_entries(
        self, courses: List[Course], core_contents: Dict[int, List[Dict]]
    ) -> Dict[int, Dict[int, Dict]]:
        workshops = (
            await self.client.async_post(
                'mod_workshop_get_workshops_by_courses', self.get_data_for_mod_entries_endpoint(courses)
            )
        ).get('workshops', [])

        result = {}
        for workshop in workshops:
            course_id = workshop.get('course', 0)
            workshop_files = workshop.get('introfiles', [])
            workshop_files += workshop.get('instructauthorsfiles', [])
            workshop_files += workshop.get('instructreviewersfiles', [])
            workshop_files += workshop.get('conclusionfiles', [])
            self.set_props_of_files(workshop_files, type='workshop_introfile')

            workshop_intro = workshop.get('intro', '')
            if workshop_intro != '':
                workshop_files.append(
                    {
                        'filename': 'Workshop intro',
                        'filepath': '/',
                        'description': workshop_intro,
                        'type': 'description',
                    }
                )

            workshop_instruct_authors = workshop.get('instructauthors', '')
            if workshop_instruct_authors != '':
                workshop_files.append(
                    {
                        'filename': 'Instructions for submission',
                        'filepath': '/',
                        'description': workshop_instruct_authors,
                        'type': 'description',
                    }
                )

            workshop_instruct_reviewers = workshop.get('instructreviewers', '')
            if workshop_instruct_reviewers != '':
                workshop_files.append(
                    {
                        'filename': 'Instructions for assessment',
                        'filepath': '/',
                        'description': workshop_instruct_reviewers,
                        'type': 'description',
                    }
                )

            workshop_conclusion = workshop.get('conclusion', '')
            if workshop_conclusion != '':
                workshop_files.append(
                    {
                        'filename': 'Conclusion',
                        'filepath': '/',
                        'description': workshop_conclusion,
                        'type': 'description',
                    }
                )

            self.add_module(
                result,
                course_id,
                workshop.get('coursemodule', 0),
                {
                    'id': workshop.get('id', 0),
                    'name': workshop.get('name', 'unnamed workshop'),
                    'files': workshop_files,
                },
            )

        await self.add_workshops_files(result)
        return result

    async def add_workshops_files(self, workshops: Dict[int, Dict[int, Dict]]):
        """
        Fetches for the workshops list the forum posts
        @param workshops: Dictionary of all workshops, indexed by courses, then module id
        """
        if not self.config.get_download_workshops():
            return

        if self.version < 2017111300:  # 3.4
            return

        await self.run_async_load_function_on_mod_entries(workshops, self.load_workshop_files)

    async def load_workshop_files(self, workshop: Dict):
        workshop_id = workshop.get('id', 0)
        data = {'workshopid': workshop_id, 'userid': self.user_id}

        try:
            submissions = (await self.client.async_post('mod_workshop_get_submissions', data)).get('submissions', [])
        except RequestRejectedError:
            logging.debug("No access rights for workshop %d", workshop_id)
            return

        try:
            assessments = (await self.client.async_post('mod_workshop_get_reviewer_assessments', data)).get(
                'assessments', []
            )
        except RequestRejectedError:
            assessments = []
        submissions += await self.run_async_collect_function_on_list(
            assessments,
            self.load_foreign_submission,
            'foreign submission',
            {'collect_id': 'submissionid', 'collect_name': 'title'},
        )

        try:
            grades = await self.client.async_post('mod_workshop_get_grades', data)
        except RequestRejectedError:
            grades = {}

        workshop_files = self._get_files_of_workshop(submissions, grades)
        workshop['files'] += workshop_files

    async def load_foreign_submission(self, assessment: Dict) -> Dict:
        # assessment_id = assessment.get('id', 0)
        # assessment_reviewer_id = assessment.get('reviewerid', 0)

        assessment_files = assessment.get('feedbackcontentfiles', [])
        assessment_files += assessment.get('feedbackattachmentfiles', [])

        feedback_author = assessment.get('feedbackauthor', '')
        if feedback_author != '':
            assessment_files.append(
                {
                    'filename': 'Feedback for the author',
                    'filepath': '/',
                    'description': feedback_author,
                    'type': 'description',
                }
            )

        feedback_reviewer = assessment.get('feedbackreviewer', '')
        if feedback_reviewer != '':
            assessment_files.append(
                {
                    'filename': 'Feedback for the reviewer',
                    'filepath': '/',
                    'description': feedback_reviewer,
                    'type': 'description',
                }
            )
        assessment_submission_id = assessment.get('submissionid', 0)
        # Get submissions of assessments
        data = {'submissionid': assessment_submission_id}
        try:
            submission = (await self.client.async_post('mod_workshop_get_submission', data)).get('submission', {})
            submission['files'] = assessment_files
            return submission
        except RequestRejectedError:
            logging.debug("No access rights for workshop submission %d", assessment_submission_id)
            return None

    def _get_files_of_workshop(self, submissions: List[Dict], grades: Dict) -> List:
        result = []

        # Grades
        assessment_long_str_grade = grades.get('assessmentlongstrgrade', '')
        if assessment_long_str_grade != '':
            result.append(
                {
                    'filename': 'Assessment grade',
                    'filepath': '/',
                    'description': assessment_long_str_grade,
                    'type': 'description',
                }
            )

        submission_long_str_grade = grades.get('submissionlongstrgrade', '')
        if submission_long_str_grade != '':
            result.append(
                {
                    'filename': 'Submission grade',
                    'filepath': '/',
                    'description': submission_long_str_grade,
                    'type': 'description',
                }
            )

        # Own and foreign submissions
        for submission in submissions:
            submission_content = submission.get('content', 0)

            filepath = f"/submissions {submission.get('id', 0)}/"

            submission_files = submission.get('contentfiles', [])
            submission_files += submission.get('attachmentfiles', [])
            self.set_props_of_files(submission_files, type='workshop_file')
            self.set_base_file_path_of_files(submission_files, filepath)

            submission_files += submission.get('files', [])  # Already pares files

            if submission_content != '':
                submission_files.append(
                    {
                        'filename': submission.get('title', 0),
                        'filepath': filepath,
                        'description': submission_content,
                        'timemodified': submission.get('timemodified', 0),
                        'type': 'description',
                    }
                )
            result += submission_files

        return result



import json
from typing import Dict, List

from requests.exceptions import RequestException

from moodle_dl.utils import SslHelper


class DiscordShooter:
    RQ_HEADER = {
        'User-Agent': (
            'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/78.0.3904.108 Safari/537.36'
        ),
        'Content-Type': 'application/json',
    }

    def __init__(self, discord_webhooks: List[str]):
        self.discord_webhooks = discord_webhooks

    def send_msg(self, text):
        self.send_data(
            {
                'content': text,
                'username': 'Moodle Notifications',
                'avatar_url': 'https://i.imgur.com/J3Pxl41.png',
            }
        )

    def send(self, embeds: List):
        self.send_data(
            {
                'embeds': embeds,
                'username': 'Moodle Notifications',
                'avatar_url': 'https://i.imgur.com/J3Pxl41.png',
            }
        )

    def send_data(self, data: Dict):

        session = SslHelper.custom_requests_session(skip_cert_verify=False, allow_insecure_ssl=False)
        for webhook_url in self.discord_webhooks:
            try:
                response = session.post(webhook_url, data=json.dumps(data), headers=self.RQ_HEADER, timeout=60)
                self._check_response_code(response)
            except RequestException as error:
                raise ConnectionError(f"Connection error: {str(error)}") from None

    @staticmethod
    def _check_response_code(response):
        # Normally Discord answers with response 204
        if response.status_code not in [200, 204, 400]:
            raise RuntimeError(
                'An unexpected error happened on'
                + " Discord's servers."
                + f' Status code: {str(response.status_code)}'
                + f'\nHeader: {response.headers}'
                + f'\nResponse: {response.text}'
            )



import re
from urllib.parse import urlparse, urlunparse

from yt_dlp.extractor.common import InfoExtractor
from yt_dlp.utils import determine_ext, float_or_none, traverse_obj


class Echo360IE(InfoExtractor):
    _VALID_URL = r'''(?x)
        https?://(?P<host>echo360\.(?:ca|net\.au|org|org\.au|org\.uk))/
        media/(?P<id>[\da-fA-F]{8}-(?:[\da-fA-F]{4}-){3}[\da-fA-F]{12})/public
    '''

    _TESTS = [
        {
            'url': 'https://echo360.org.uk/media/1d8392aa-a3e7-4e78-94cf-b6532c27208c/public',
            'info_dict': {
                'id': '3c7ae6e0-fa19-432d-aa21-c283b4276f2a',
                'ext': 'mp4',
                'title': '3-4 Force + moment + mechanics.mp4',
                'duration': 4731.888,
            },
            'params': {'skip_download': 'm3u8'},
        },
        {
            'url': 'https://echo360.net.au/media/f04960a9-2efc-4b63-87b5-72e629081d15/public',
            'info_dict': {
                'id': '6098a147-2d65-40f3-b9e9-a0204afe450c',
                'ext': 'mp4',
                'title': 'EXSC634_Online_Workshop_Week_4.mp4',
                'duration': 6659.72,
            },
            'params': {'skip_download': 'm3u8'},
        },
    ]

    def _call_api(self, host, video_id, media_id, session_token, **kwargs):
        return self._download_json(
            f'https://{host}/api/ui/echoplayer/public-links/{video_id}/media/{media_id}/player-properties',
            video_id,
            headers={'Authorization': f'Bearer {session_token}'},
            **kwargs,
        )

    def _replace_url_query(self, url, query_string):
        if query_string is not None:
            return urlunparse(urlparse(url)._replace(query=query_string))
        return url

    def _get_query_string(self, uri, query_strings):
        uri_base = urlparse(uri)._replace(query='', fragment='').geturl()
        for query_string in query_strings:
            try:
                if re.match(query_string['uriPattern'], uri_base):
                    return query_string['queryString']
            except re.error as re_error:
                self.report_warning(f'Error in query string pattern `{re_error.pattern}`: {re_error.msg}')
        return None

    def _parse_mediapackage(self, video):
        video_id = video['playableAudioVideo']['mediaId']
        query_strings = traverse_obj(video, ('sourceQueryStrings', 'queryStrings')) or []

        formats = []
        for track in traverse_obj(video, ('playableAudioVideo', 'playableMedias', ...)):
            href = track.get('uri')
            if href is None:
                continue
            href = self._replace_url_query(href, self._get_query_string(href, query_strings))
            if track.get('isHls') or determine_ext(href, None) == 'm3u8':
                hls_formats = self._extract_m3u8_formats(
                    href, video_id, live=track.get('isLive'), m3u8_id='hls', entry_protocol='m3u8_native', fatal=False
                )

                for hls_format in hls_formats:
                    query_string = self._get_query_string(hls_format['url'], query_strings)
                    hls_format['extra_param_to_segment_url'] = query_string
                    hls_format['url'] = self._replace_url_query(hls_format['url'], query_string)

                formats.extend(hls_formats)

        return {
            'id': video_id,
            'formats': formats,
            'title': video.get('mediaName'),
            'duration': float_or_none(
                self._search_regex(
                    r'PT([\d.]+)S',
                    traverse_obj(video, ('playableAudioVideo', 'duration')),
                    'video duration',
                    fatal=False,
                )
            ),
        }

    def _real_extract(self, url):
        host, video_id = self._match_valid_url(url).group('host', 'id')
        webpage = self._download_webpage(url, video_id)

        player_config = self._search_json(
            r'Echo\["mediaPlayerBootstrapApp"\]\("',
            webpage,
            'player config',
            video_id,
            transform_source=lambda x: x.replace(R'\"', '"'),
        )

        urlh = self._request_webpage(
            f'https://{host}/api/ui/sessions/{player_config["sessionId"]}',
            video_id,
            'Open video session',
            'Unable to open video session',
        )

        return self._parse_mediapackage(
            self._call_api(
                host,
                player_config.get('shareLinkId') or player_config['publicLinkId'],
                player_config['mediaId'],
                urlh.headers['Token'],
            )['data']
        )



from __future__ import unicode_literals

import re

from yt_dlp.compat import compat_parse_qs
from yt_dlp.extractor.common import InfoExtractor
from yt_dlp.utils import (
    ExtractorError,
    int_or_none,
    lowercase_escape,
    try_get,
    update_url_query,
)

from moodle_dl.utils import determine_ext


class GoogleDriveIE(InfoExtractor):
    _VALID_URL = r'''(?x)
                        https?://
                            (?:
                                (?:docs|drive)\.google\.com/
                                (?:
                                    (?:uc|open)\?.*?id=|
                                    file/d/
                                )|
                                video\.google\.com/get_player\?.*?docid=
                            )
                            (?P<id>[a-zA-Z0-9_-]{28,})
                    '''
    _TESTS = [
        {
            'url': 'https://drive.google.com/file/d/0ByeS4oOUV-49Zzh4R1J6R09zazQ/edit?pli=1',
            'md5': '5c602afbbf2c1db91831f5d82f678554',
            'info_dict': {
                'id': '0ByeS4oOUV-49Zzh4R1J6R09zazQ',
                'ext': 'mp4',
                'title': 'Big Buck Bunny.mp4',
                'duration': 45,
            },
        },
        {
            # video can't be watched anonymously due to view count limit reached,
            # but can be downloaded (see https://github.com/ytdl-org/youtube-dl/issues/14046)
            'url': 'https://drive.google.com/file/d/0B-vUyvmDLdWDcEt4WjBqcmI2XzQ/view',
            'only_matching': True,
        },
        {
            # video id is longer than 28 characters
            'url': 'https://drive.google.com/file/d/1ENcQ_jeCuj7y19s66_Ou9dRP4GKGsodiDQ/edit',
            'only_matching': True,
        },
        {
            'url': 'https://drive.google.com/open?id=0B2fjwgkl1A_CX083Tkowdmt6d28',
            'only_matching': True,
        },
        {
            'url': 'https://drive.google.com/uc?id=0B2fjwgkl1A_CX083Tkowdmt6d28',
            'only_matching': True,
        },
    ]
    _FORMATS_EXT = {
        '5': 'flv',
        '6': 'flv',
        '13': '3gp',
        '17': '3gp',
        '18': 'mp4',
        '22': 'mp4',
        '34': 'flv',
        '35': 'flv',
        '36': '3gp',
        '37': 'mp4',
        '38': 'mp4',
        '43': 'webm',
        '44': 'webm',
        '45': 'webm',
        '46': 'webm',
        '59': 'mp4',
    }
    _BASE_URL_CAPTIONS = 'https://drive.google.com/timedtext'
    _CAPTIONS_ENTRY_TAG = {
        'subtitles': 'track',
        'automatic_captions': 'target',
    }
    _caption_formats_ext = []
    _captions_xml = None

    @staticmethod
    def _extract_url(webpage):
        mobj = re.search(
            r'''(?x)<iframe[^>]+src="https?://(?:video\.google\.com
            /get_player\?.*?docid=|(?:docs|drive)\.google\.com/file/d/)(?P<id>[a-zA-Z0-9_-]{28,})''',
            webpage,
        )
        if mobj:
            return f"https://drive.google.com/file/d/{mobj.group('id')}"

    def _download_subtitles_xml(self, video_id, subtitles_id, hl):
        if self._captions_xml:
            return
        self._captions_xml = self._download_xml(
            self._BASE_URL_CAPTIONS,
            video_id,
            query={
                'id': video_id,
                'vid': subtitles_id,
                'hl': hl,
                'v': video_id,
                'type': 'list',
                'tlangs': '1',
                'fmts': '1',
                'vssids': '1',
            },
            note='Downloading subtitles XML',
            errnote='Unable to download subtitles XML',
            fatal=False,
        )
        if self._captions_xml:
            for f in self._captions_xml.findall('format'):
                if f.attrib.get('fmt_code') and not f.attrib.get('default'):
                    self._caption_formats_ext.append(f.attrib['fmt_code'])

    def _get_captions_by_type(self, video_id, subtitles_id, caption_type, origin_lang_code=None):
        if not subtitles_id or not caption_type:
            return
        captions = {}
        for caption_entry in self._captions_xml.findall(self._CAPTIONS_ENTRY_TAG[caption_type]):
            caption_lang_code = caption_entry.attrib.get('lang_code')
            if not caption_lang_code:
                continue
            caption_format_data = []
            for caption_format in self._caption_formats_ext:
                query = {
                    'vid': subtitles_id,
                    'v': video_id,
                    'fmt': caption_format,
                    'lang': (caption_lang_code if origin_lang_code is None else origin_lang_code),
                    'type': 'track',
                    'name': '',
                    'kind': '',
                }
                if origin_lang_code is not None:
                    query.update({'tlang': caption_lang_code})
                caption_format_data.append(
                    {
                        'url': update_url_query(self._BASE_URL_CAPTIONS, query),
                        'ext': caption_format,
                    }
                )
            captions[caption_lang_code] = caption_format_data
        return captions

    def _get_subtitles(self, *args, **kwargs):
        if len(args) != 3:
            return
        video_id = args[0]
        subtitles_id = args[1]
        hl = args[2]
        self._download_subtitles_xml(video_id, subtitles_id, hl)
        if not self._captions_xml:
            return
        return self._get_captions_by_type(video_id, subtitles_id, 'subtitles')

    def _get_automatic_captions(self, *args, **kwargs):
        if len(args) != 3:
            return
        video_id = args[0]
        subtitles_id = args[1]
        hl = args[2]
        self._download_subtitles_xml(video_id, subtitles_id, hl)
        if not self._captions_xml:
            return
        track = self._captions_xml.find('track')
        if track is None:
            return
        origin_lang_code = track.attrib.get('lang_code')
        if not origin_lang_code:
            return
        return self._get_captions_by_type(video_id, subtitles_id, 'automatic_captions', origin_lang_code)

    def _real_extract(self, url):
        video_id = self._match_id(url)

        # First try the new method
        video_info = compat_parse_qs(
            self._download_webpage('https://drive.google.com/get_video_info', video_id, query={'docid': video_id})
        )

        def get_value(key):
            return try_get(video_info, lambda x: x[key][0])

        reason = get_value('reason')
        title = get_value('title')

        use_old_webpage = False
        if not title and reason:
            use_old_webpage = True
            webpage = self._download_webpage(f'http://docs.google.com/file/d/{video_id}', video_id)

            title = self._search_regex(
                r'"title"\s*,\s*"([^"]+)', webpage, 'title', default=None
            ) or self._og_search_title(webpage)
            duration = int_or_none(
                self._search_regex(r'"length_seconds"\s*,\s*"([^"]+)', webpage, 'length seconds', default=None)
            )

            formats = []
            fmt_stream_map = self._search_regex(
                r'"fmt_stream_map"\s*,\s*"([^"]+)', webpage, 'fmt stream map', default=''
            ).split(',')
            fmt_list = self._search_regex(r'"fmt_list"\s*,\s*"([^"]+)', webpage, 'fmt_list', default='').split(',')

            hl = self._search_regex(r'"hl"\s*,\s*"([^"]+)', webpage, 'hl', default=None)
            subtitles_id = None
            ttsurl = self._search_regex(r'"ttsurl"\s*,\s*"([^"]+)', webpage, 'ttsurl', default=None)

            thumbnail_url = self._og_search_thumbnail(webpage, default=None)
        else:
            formats = []
            fmt_stream_map = (get_value('fmt_stream_map') or '').split(',')
            fmt_list = (get_value('fmt_list') or '').split(',')

            duration = int_or_none(get_value('length_seconds'))

            hl = get_value('hl')
            subtitles_id = None
            ttsurl = get_value('ttsurl')
            thumbnail_url = 'https://drive.google.com/thumbnail?id=' + video_id

        if ttsurl:
            # the video Id for subtitles will be the last value in the ttsurl
            # query string
            subtitles_id = ttsurl.encode('utf-8').decode('unicode_escape').split('=')[-1]

        if fmt_stream_map and fmt_list:
            resolutions = {}
            for fmt in fmt_list:
                mobj = re.search(r'^(?P<format_id>\d+)/(?P<width>\d+)[xX](?P<height>\d+)', fmt)
                if mobj:
                    resolutions[mobj.group('format_id')] = (int(mobj.group('width')), int(mobj.group('height')))

            for fmt_stream in fmt_stream_map:
                fmt_stream_split = fmt_stream.split('|')
                if len(fmt_stream_split) < 2:
                    continue
                format_id, format_url = fmt_stream_split[:2]
                f = {
                    'url': lowercase_escape(format_url),
                    'format_id': format_id,
                    'ext': self._FORMATS_EXT[format_id],
                }
                resolution = resolutions.get(format_id)
                if resolution:
                    f.update(
                        {
                            'width': resolution[0],
                            'height': resolution[1],
                        }
                    )
                formats.append(f)

        source_url = update_url_query(
            'https://drive.google.com/uc',
            {
                'id': video_id,
                'export': 'download',
            },
        )

        def request_source_file(source_url, kind):
            return self._request_webpage(
                source_url,
                video_id,
                note=f'Requesting {kind} file',
                errnote=f'Unable to request {kind} file',
                fatal=False,
            )

        urlh = request_source_file(source_url, 'source')
        if urlh:

            def add_source_format(urlh):
                formats.append(
                    {
                        # Use redirect URLs as download URLs in order to calculate
                        # correct cookies in _calc_cookies.
                        # Using original URLs may result in redirect loop due to
                        # google.com's cookies mistakenly used for googleusercontent.com
                        # redirect URLs (see #23919).
                        'url': urlh.geturl(),
                        'ext': determine_ext(title, 'mp4').lower(),
                        'format_id': 'source',
                        'quality': 1,
                    }
                )

            if urlh.headers.get('Content-Disposition'):
                add_source_format(urlh)
            else:
                confirmation_webpage = self._webpage_read_content(
                    urlh,
                    url,
                    video_id,
                    note='Downloading confirmation page',
                    errnote='Unable to confirm download',
                    fatal=False,
                )
                if confirmation_webpage:
                    confirm = self._search_regex(
                        r'confirm=([^&"\']+)', confirmation_webpage, 'confirmation code', fatal=False
                    )
                    if confirm:
                        confirmed_source_url = update_url_query(
                            source_url,
                            {
                                'confirm': confirm,
                            },
                        )
                        urlh = request_source_file(confirmed_source_url, 'confirmed source')
                        if urlh and urlh.headers.get('Content-Disposition'):
                            add_source_format(urlh)

        if not use_old_webpage and not formats and reason:
            raise ExtractorError(reason, expected=True)
        elif use_old_webpage and not formats:
            reason = self._search_regex(r'"reason"\s*,\s*"([^"]+)', webpage, 'reason', default=None)
            if reason:
                raise ExtractorError(reason, expected=True)

        self._sort_formats(formats)

        return {
            'id': video_id,
            'title': title,
            'thumbnail': thumbnail_url,
            'duration': duration,
            'formats': formats,
            'subtitles': self.extract_subtitles(video_id, subtitles_id, hl),
            'automatic_captions': self.extract_automatic_captions(video_id, subtitles_id, hl),
        }



import logging
import traceback
from typing import List

from moodle_dl.downloader.task import Task
from moodle_dl.notifications.discord.discord_formatter import DiscordFormatter as DF
from moodle_dl.notifications.discord.discord_shooter import DiscordShooter
from moodle_dl.notifications.notification_service import NotificationService
from moodle_dl.types import Course


class DiscordService(NotificationService):

    def _is_configured(self) -> bool:
        # Checks if the sending of Discord messages has been configured.
        try:
            self.config.get_property('discord')
            return True
        except ValueError:
            logging.debug('Discord webhook notifications not configured, skipping.')
            return False

    def _send_embeds(self, embeds: []):
        """
        Sends a Discord webhook notification
        """
        if not self._is_configured():
            return

        discord_cfg = self.config.get_property('discord')

        logging.info('Sending Notification via Discord webhooks...')

        discord_shooter = DiscordShooter(discord_cfg['webhook_urls'])

        try:
            discord_shooter.send(embeds)
        except BaseException as e:
            logging.error('While sending notification:\n%s', traceback.format_exc(), extra={'exception': e})
            raise e  # to be properly notified via Sentry

    def notify_about_changes_in_moodle(self, changes: [Course]) -> None:
        """
        Sends out a notification about the downloaded changes.
        @param changes: A list of changed courses with changed files.
        """
        if not self._is_configured():
            return

        messages = DF.create_full_moodle_diff_messages(changes, self.config.get_moodle_URL().url_base)

        self._send_embeds(messages)

    def notify_about_error(self, error_description: str) -> None:
        # Not yet implemented
        pass

    def notify_about_failed_downloads(self, failed_downloads: List[Task]) -> None:
        # Not yet implemented
        pass



from moodle_dl.types import Course


class DiscordFormatter:
    @classmethod
    def make_bold(cls, string: str) -> str:
        """
        Makes a string bold in a telegram message
        """
        return '**' + string + '**'

    @classmethod
    def create_full_moodle_diff_messages(cls, changed_courses: [Course], moodle_url_base) -> [str]:
        """
        Creates Discord messages with all changed files. This includes new,
        modified and deleted files. Files that have changed since the last message.

        @param changed_courses: A list of all courses with their modified files.
        @returns embeds to send
        """

        embeds = []

        for course in changed_courses:
            new_embed = {
                'author': {
                    'name': course.fullname,
                    'url': f"{moodle_url_base}course/view.php?id={course.id}",
                    'icon_url': 'https://i.imgur.com/Bt5TFIA.png',
                },
                'fields': [],
            }

            for file in course.files:
                saved_to_path = file.saved_to
                if file.new_file is not None:
                    saved_to_path = file.new_file.saved_to

                field_name = 'Added'
                new_embed['color'] = '7268279'  # emerald green

                if file.modified:
                    field_name = 'Modified'
                    new_embed['color'] = '16628340'  # orange
                elif file.moved:
                    field_name = 'Moved'
                    new_embed['color'] = '8246268'  # sky blue
                elif file.deleted:
                    field_name = 'Deleted'
                    new_embed['color'] = '16622767'  # rose

                value = f"• {saved_to_path.replace(f'{course.fullname}/', '')}"

                for field in new_embed['fields']:
                    if field['name'] == field_name:
                        field['value'] += f"\n{value}"
                        break
                else:
                    new_embed['fields'].append({'name': field_name, 'value': value})

            for field in new_embed['fields']:
                value = field['value']
                field['value'] = value[:1021] + '...' if len(value) > 1024 else value
            embeds.append(new_embed)

        return embeds



import html
import json
import urllib.parse
from typing import Dict, List

from moodle_dl.config import ConfigHelper
from moodle_dl.moodle.mods import MoodleMod
from moodle_dl.types import Course, File


class BookMod(MoodleMod):
    MOD_NAME = 'book'
    MOD_PLURAL_NAME = 'books'
    MOD_MIN_VERSION = 2015111600  # 3.0

    @classmethod
    def download_condition(cls, config: ConfigHelper, file: File) -> bool:
        return config.get_download_books() or (not (file.module_modname.endswith(cls.MOD_NAME) and file.deleted))

    async def real_fetch_mod_entries(
        self, courses: List[Course], core_contents: Dict[int, List[Dict]]
    ) -> Dict[int, Dict[int, Dict]]:

        result = {}
        if not self.config.get_download_books():
            return result

        books = (
            await self.client.async_post(
                'mod_book_get_books_by_courses', self.get_data_for_mod_entries_endpoint(courses)
            )
        ).get('books', [])

        for book in books:
            course_id = book.get('course', 0)
            module_id = book.get('coursemodule', 0)
            book_name = book.get('name', 'unnamed book')

            book_files = book.get('introfiles', [])
            self.set_props_of_files(book_files, type='book_file')

            book_intro = book.get('intro', '')
            if book_intro != '':
                book_files.append(
                    {
                        'filename': 'Book intro',
                        'filepath': '/',
                        'description': book_intro,
                        'type': 'description',
                    }
                )

            book_contents = self.get_module_in_core_contents(course_id, module_id, core_contents).get('contents', [])
            if len(book_contents) > 1:
                book_files += book_contents[1:]

            if len(book_contents) > 0:
                # Generate Table of Contents
                book_toc = json.loads(book_contents[0].get('content', ''))

                toc_html = '''
<!DOCTYPE html>
<html>
    <head>
        <style>
            ol {
                counter-reset: item
            }
            li {
                display: block
            }
            li:before {
                content: counters(item, ".")" ";
                counter-increment: item
            }
        </style>
    </head>
    <body>
        '''
                toc_html += self.create_ordered_index(book_toc)
                toc_html += '''
    </body>
</html>
                '''

                book_files.append(
                    {
                        'filename': 'Table of Contents',
                        'filepath': '/',
                        'timemodified': book.get('timemodified', 0),
                        'html': toc_html,
                        'type': 'html',
                        'no_search_for_urls': True,
                        'filesize': len(toc_html),
                    }
                )

            self.add_module(
                result,
                course_id,
                module_id,
                {
                    'id': book.get('id', 0),
                    'name': book_name,
                    'files': book_files,
                },
            )

        return result

    @staticmethod
    def create_ordered_index(items: List[Dict]) -> str:
        result = '<ol>\n'
        for entry in items:
            chapter_title = html.escape(entry.get("title", "untitled"))
            chapter_href = urllib.parse.quote(entry.get("href", "#failed"))
            result += f'<li><a title="{chapter_title}" href="{chapter_href}">{chapter_title}</a></li>\n'
            subitems = entry.get('subitems', [])
            if len(subitems) > 0:
                result += BookMod.create_ordered_index(subitems)

        result += '</ol>'
        return result



import logging
from datetime import datetime
from typing import Dict, List

from moodle_dl.config import ConfigHelper
from moodle_dl.moodle.mods import MoodleMod
from moodle_dl.types import Course, File
from moodle_dl.utils import PathTools as PT


class ForumMod(MoodleMod):
    MOD_NAME = 'forum'
    MOD_PLURAL_NAME = 'forums'
    MOD_MIN_VERSION = 2013051400  # 2.5

    @classmethod
    def download_condition(cls, config: ConfigHelper, file: File) -> bool:
        # TODO: Add download condition, currently forums get filtered on API Call, and are not deleted at all
        return True

    async def real_fetch_mod_entries(
        self, courses: List[Course], core_contents: Dict[int, List[Dict]]
    ) -> Dict[int, Dict[int, Dict]]:
        forums = await self.client.async_post(
            'mod_forum_get_forums_by_courses', self.get_data_for_mod_entries_endpoint(courses)
        )

        result = {}
        for forum in forums:
            course_id = forum.get('course', 0)
            forum_module_id = forum.get('cmid', 0)
            forum_files = forum.get('introfiles', [])
            self.set_props_of_files(forum_files, type='forum_introfile')

            forum_intro = forum.get('intro', '')
            if forum_intro != '':
                forum_files.append(
                    {
                        'filename': 'Forum intro',
                        'filepath': '/',
                        'description': forum_intro,
                        'type': 'description',
                    }
                )

            self.add_module(
                result,
                course_id,
                forum_module_id,
                {
                    'id': forum.get('id', 0),
                    'name': forum.get('name', 'forum'),
                    'files': forum_files,
                    '_cmid': forum_module_id,
                },
            )

        await self.add_forum_posts(result)
        return result

    async def add_forum_posts(self, forums: Dict[int, Dict[int, Dict]]):
        """
        Fetches for the forums list the forum posts
        @param forums: Dictionary of all forums, indexed by courses, then module id
        """
        if not self.config.get_download_forums():
            return

        if self.version < 2014111000:  # 2.8
            return

        await self.run_async_load_function_on_mod_entries(forums, self.load_latest_discussions)

    async def load_latest_discussions(self, forum: Dict):
        "Adds the discussions that needs to be updated to the forum dict"
        page_num = 0
        last_timestamp = self.last_timestamps.get(self.MOD_NAME, {}).get(forum.get('_cmid', 0), 0)
        latest_discussions = []
        done = False
        while not done:
            data = {
                'forumid': forum.get('id', 0),
                'perpage': 10,
                'page': page_num,
            }

            if self.version >= 2019052000:  # 3.7
                discussions_result = await self.client.async_post('mod_forum_get_forum_discussions', data)
            else:
                discussions_result = await self.client.async_post('mod_forum_get_forum_discussions_paginated', data)

            logging.debug(
                'Loaded %(mod_name)s page %(page_num)d of "%(forum_name)s"',
                {'mod_name': self.MOD_NAME, 'page_num': page_num, 'forum_name': forum.get('name', '')},
            )

            discussions = discussions_result.get('discussions', [])

            if len(discussions) == 0:
                done = True
                break

            for discussion in discussions:
                time_modified = discussion.get('timemodified', 0)
                if discussion.get('modified', 0) > time_modified:
                    time_modified = discussion.get('modified', 0)

                if last_timestamp < time_modified:
                    latest_discussions.append(
                        {
                            'subject': discussion.get('subject', ''),
                            'timemodified': time_modified,
                            'discussion_id': discussion.get('discussion', 0),
                            'created': discussion.get('created', 0),
                        }
                    )
                else:
                    done = True
                    break
            page_num += 1

        forum['files'] += await self.run_async_collect_function_on_list(
            latest_discussions,
            self.load_files_of_discussion,
            'discussion',
            {'collect_id': 'discussion_id', 'collect_name': 'subject'},
        )

    async def load_files_of_discussion(self, discussion: Dict) -> List[Dict]:
        result = []

        data = {
            'discussionid': discussion.get('discussion_id', 0),
            'sortby': 'modified',
            'sortdirection': 'ASC',
        }
        if self.version >= 2019052000:  # 3.7
            posts = (await self.client.async_post('mod_forum_get_discussion_posts', data)).get('posts', [])
        else:
            posts = (await self.client.async_post('mod_forum_get_forum_discussion_posts', data)).get('posts', [])

        for post in posts:
            post_message = post.get('message', '') or ''

            post_files = post.get('attachments', [])
            if self.version >= 2019052000:  # 3.7
                post_parent = post.get('parentid', 0)
                post_user_fullname = post.get('author', {}).get('fullname', None) or 'Unknown'
                post_modified = post.get('timecreated', 0)
                for post_file in post_files:
                    # New return structure uses url instead of fileurl
                    post_file['fileurl'] = post_file.get('url', '')
                    # And also do return normal URLs instead of webservice URLs
                    if post_file['fileurl'].find('/webservice/') < 0:
                        post_file['fileurl'] = post_file['fileurl'].replace(
                            '/pluginfile.php/', '/webservice/pluginfile.php/'
                        )
            else:
                post_parent = post.get('parent', 0)
                post_user_fullname = post.get('userfullname', '') or 'Unknown'
                post_modified = post.get('modified', 0)

                # Also add legacy inline files from messageinlinefiles attribute
                self.add_legacy_inline_files(post.get('messageinlinefiles', []), post_file)

            post_filename = PT.to_valid_name(f"[{post.get('id', 0)}] " + post_user_fullname, is_file=False)
            if post_parent is not None and post_parent != 0:
                post_filename = PT.to_valid_name(
                    post_filename + ' response to [' + str(post_parent) + ']', is_file=False
                )

            post_path = PT.to_valid_name(
                datetime.utcfromtimestamp(discussion.get('created', 0)).strftime('%y-%m-%d')
                + ' '
                + discussion.get('subject', ''),
                is_file=False,
            )

            result.append(
                {
                    'filename': post_filename,
                    'filepath': post_path,
                    'timemodified': post_modified,
                    'description': post_message,
                    'type': 'description',
                }
            )

            self.set_props_of_files(post_files, type='forum_file')
            self.set_base_file_path_of_files(post_files, post_path)

            result.extend(post_files)

        return result

    def add_legacy_inline_files(self, inline_files: List, post_files: List):
        for inline_file in inline_files:
            new_inline_file = True
            for attachment in post_files:
                if attachment.get('fileurl', '').replace('attachment', 'post') == inline_file.get('fileurl', ''):
                    if (
                        attachment.get('filesize', 0) == inline_file.get('filesize', 0)
                        # we assume that inline attachments can have different timestamps than the actual
                        # attachment. However, they are still the same file.
                        and attachment.get('filename', '') == inline_file.get('filename', '')
                    ):
                        new_inline_file = False
                        break
            if new_inline_file:
                post_files.append(inline_file)



import asyncio
import functools
import logging
import os
import posixpath
import re
import shlex
import shutil
import subprocess
import time
import traceback
import urllib
import urllib.parse as urlparse
from concurrent.futures import ThreadPoolExecutor
from email.utils import unquote
from io import StringIO
from pathlib import Path
from typing import Callable, Dict
from urllib.error import ContentTooShortError

import aiofiles
import aiohttp
import html2text
import yt_dlp

from moodle_dl.downloader.extractors import add_additional_extractors
from moodle_dl.types import (
    Course,
    DlEvent,
    DownloadOptions,
    File,
    HeadInfo,
    TaskState,
    TaskStatus,
)
from moodle_dl.utils import LINK_TEMPLATES, MoodleDLCookieJar
from moodle_dl.utils import PathTools as PT
from moodle_dl.utils import (
    SslHelper,
    Timer,
    convert_to_aiohttp_cookie_jar,
    format_bytes,
    format_seconds,
    timeconvert,
)


class Task:
    "Task is responsible to download or create a file"
    CHUNK_SIZE = 102400  # default: 1024 * 100 = 100kb; will be overwritten with download_chunk_size
    MAX_DL_RETRIES = 3

    RQ_HEADER = {
        'User-Agent': (
            'Mozilla/5.0 (Linux; Android 7.1.1; Moto G Play Build/NPIS26.48-43-2; wv) AppleWebKit/537.36'
            + ' (KHTML, like Gecko) Version/4.0 Chrome/71.0.3578.99 Mobile Safari/537.36 MoodleMobile'
        ),
        'Content-Type': 'application/x-www-form-urlencoded',
    }

    def __init__(
        self,
        task_id: int,
        file: File,
        course: Course,
        options: DownloadOptions,
        thread_pool: ThreadPoolExecutor,
        callback: Callable[[], None],
    ):
        self.task_id = task_id
        self.file = file
        self.course = course
        self.opts = options
        self.thread_pool = thread_pool
        self.callback = callback

        self.destination = self.gen_path(options.download_path, course, file)
        self.filename = PT.to_valid_name(self.file.content_filename, is_file=True)
        self.status = TaskStatus()

    @staticmethod
    def gen_path(storage_path: str, course: Course, file: File):
        "Generate the directory path where a file should be stored"
        course_name = course.fullname
        if course.overwrite_name_with is not None:
            course_name = course.overwrite_name_with

        # TODO: Move this out of the downloader
        # if a flat path is requested
        if not course.create_directory_structure:
            return PT.flat_path_of_file(storage_path, course_name, file.content_filepath)

        # TODO: Get mod names automated; all mods should be in a sub-folder
        # If the file is located in a folder or in an assignment,
        # it should be saved in a sub-folder (with the name of the module).
        if file.module_modname.endswith(
            ('assign', 'book', 'data', 'folder', 'forum', 'lesson', 'page', 'quiz', 'workshop')
        ):
            return PT.path_of_file_in_module(
                storage_path, course_name, file.section_name, file.module_name, file.content_filepath
            )
        return PT.path_of_file(storage_path, course_name, file.section_name, file.content_filepath)

    def add_token_to_url(self, url: str) -> str:
        """
        Adds the Moodle token to a URL
        @param url: The URL to that the token should be added.
        @return: The URL with the token.
        """
        url_parts = list(urlparse.urlparse(url))
        query = dict(urlparse.parse_qsl(url_parts[4]))
        query.update({'token': self.opts.token})
        url_parts[4] = urlparse.urlencode(query)
        return urlparse.urlunparse(url_parts)

    def create_target_file(self, target_path: str) -> str:
        """
        Rename target_path if necessary to a unused filename and touch the target_path
        @return: Path to the touched target file
        """
        target_path = PT.get_unused_file_path(target_path)
        PT.touch_file(target_path)
        return target_path

    def rename_old_file(self) -> bool:
        """
        Try to rename an existing modified file. Add the extension '_old' to the filename if possible.
        @return: True on success
        """
        if self.file.old_file is None:
            return False

        old_path = self.file.old_file.saved_to
        if not os.path.exists(old_path):
            return False

        logging.debug('[%d] Renaming old file', self.task_id)

        destination, filename, file_extension = PT.get_path_parts(old_path)
        new_filename = f'{filename}_old.{file_extension}'
        new_path = PT.get_unused_file_path(PT.make_path(destination, new_filename))

        try:
            shutil.move(old_path, new_path)
            self.file.old_file.saved_to = new_path
        except OSError:
            logging.warning('[%d] Failed to renaming old file %r to %r', self.task_id, old_path, new_path)
            return False

        return True

    class YtLogger:
        "logger for yt-dlp"

        def __init__(self, task):
            self.task = task
            self.task_id = task.task_id

        def clean_msg(self, msg: str) -> str:
            msg = msg.replace('\n', '')
            msg = msg.replace('\r', '')
            msg = msg.replace('\033[K', '')
            msg = msg.replace('\033[0;31m', '')
            msg = msg.replace('\033[0m', '')
            msg = re.sub('token=([a-zA-Z0-9]+)', 'censored_sensitive_data', msg)

            return msg

        def debug(self, msg):
            if msg.find('ETA') >= 0:
                return
            msg = self.clean_msg(msg)
            logging.debug('[%d] yt-dlp Debug: %s', self.task_id, msg)

        def warning(self, msg):
            msg = self.clean_msg(msg)
            if (msg.find('Falling back on generic information extractor')) >= 0 or (
                msg.find('Forcing generic information extractor') >= 0
            ):
                self.task.status.yt_dlp_used_generic_extractor = True
                logging.debug('[%d] yt-dlp Warning: %s', self.task_id, msg)
                return
            if msg.find('Requested formats are incompatible for merge') >= 0:
                logging.debug('[%d] yt-dlp Warning: %s', self.task_id, msg)
                return
            logging.warning('[%d] yt-dlp Warning: %s', self.task_id, msg)

        def error(self, msg):
            msg = self.clean_msg(msg)
            if msg.find('Unsupported URL') >= 0:
                logging.debug('[%d] yt-dlp Error: %s', self.task_id, msg)
                return
            if msg.find('no suitable InfoExtractor') >= 0:
                logging.debug('[%d] yt-dlp Error: %s', self.task_id, msg)
                return
            # This is a critical error, with high probability the link can be downloaded at a later time.
            logging.error('[%d] yt-dlp Error: %s', self.task_id, msg)
            self.task.status.yt_dlp_failed_with_error = True

    def yt_hook(self, data: Dict):
        """
        @param data: a dictionary with the entries
            * status: One of "downloading", "error", or "finished". Check this first and ignore unknown values.
            * info_dict: The extracted info_dict

            If status is one of "downloading", or "finished", the following properties may also be present:
            * filename: The final filename (always present)
            * tmpfilename: The filename we're currently writing to
            * downloaded_bytes: Bytes on disk
            * total_bytes: Size of the whole file, None if unknown
            * total_bytes_estimate: Guess of the eventual file size, None if unavailable.
            * elapsed: The number of seconds since download started.
            * eta: The estimated time in seconds, None if unknown
            * speed: The download speed in bytes/second, None if unknown
            * fragment_index: The counter of the currently downloaded video fragment.
            * fragment_count: The number of fragments (= individual files that will be merged)

            Progress hooks are guaranteed to be called at least once
            (with status "finished") if the download is successful.
        """
        if data['status'] == 'error':
            return

        tmp_file_name = data.get('tmpfilename')
        if tmp_file_name is None or tmp_file_name == '':
            # for some unknown reason sometimes the filename is missing
            return
        content_length = data.get('total_bytes', 0) or 0
        if content_length <= 0:
            total_bytes_estimate = data.get('total_bytes_estimate', 0) or 0
            content_length = total_bytes_estimate
        bytes_received_total = data.get('downloaded_bytes', 0) or 0

        self.status.yt_dlp_current_file = tmp_file_name
        self.report_yt_dlp_content_length(content_length, tmp_file_name)
        self.report_yt_dlp_received_bytes(bytes_received_total, tmp_file_name)

    def report_yt_dlp_content_length(self, content_length: int, file_name: str):
        if file_name not in self.status.yt_dlp_total_size_per_file:
            self.status.yt_dlp_total_size_per_file[file_name] = content_length
            self.status.external_total_size += content_length
            self.callback(DlEvent.TOTAL_SIZE, self, content_length=content_length)
            return
        old_content_length = self.status.yt_dlp_total_size_per_file[file_name]
        if old_content_length != content_length:
            content_length_diff = content_length - old_content_length
            self.status.external_total_size += content_length_diff
            self.callback(DlEvent.TOTAL_SIZE_UPDATE, self, content_length_diff=content_length_diff)
            self.status.yt_dlp_total_size_per_file[file_name] = content_length

    def report_yt_dlp_received_bytes(self, bytes_received_total: int, file_name: str):
        if file_name not in self.status.yt_dlp_bytes_downloaded_per_file:
            self.status.yt_dlp_bytes_downloaded_per_file[file_name] = bytes_received_total
            self.status.bytes_downloaded += bytes_received_total
            self.callback(DlEvent.RECEIVED, self, bytes_received=bytes_received_total)
            return
        old_bytes_downloaded = self.status.yt_dlp_bytes_downloaded_per_file[file_name]
        bytes_received = bytes_received_total - old_bytes_downloaded
        if bytes_received > 0:
            self.status.yt_dlp_bytes_downloaded_per_file[file_name] = bytes_received_total
            self.status.bytes_downloaded += bytes_received
            self.callback(DlEvent.RECEIVED, self, bytes_received=bytes_received)
        elif bytes_received < 0:
            logging.debug('Calculation error in report_yt_dlp_received_bytes')

    def yt_hook_after_move(self, final_filename: str):
        """
        Get called as the final step for each video file, after all postprocessors have been called.
        @param final_filename: The filename will be passed as the only argument.
        """
        rel_pos = final_filename.find(self.destination)
        if rel_pos >= 0:
            final_filename = final_filename[rel_pos:]
        self.file.saved_to = final_filename

    def is_blocked_for_yt_dlp(self, url: str):
        url_parsed = urlparse.urlparse(url)
        # Do not download whole YT channels
        if url_parsed.hostname.endswith('youtube.com') and url_parsed.path.startswith('/channel/'):
            return True
        return False

    def set_utime(self, last_modified_header: str = None):
        """
        Sets the last modified time of the downloaded file
        Modified time will be set based on the given last_modified value or the moodle file attribute timemodified
        Access time will always be set to now

        @param last_modified_header: The last_modified header from the Webpage. Defaults to None.
        """
        if not os.path.isfile(self.file.saved_to):
            return
        try:
            if last_modified_header is not None:
                server_modified_time = timeconvert(last_modified_header)
                if server_modified_time is not None and server_modified_time > 0:
                    os.utime(self.file.saved_to, (time.time(), server_modified_time))
                    return

            if self.file.content_timemodified is not None and self.file.content_timemodified > 0:
                os.utime(self.file.saved_to, (time.time(), self.file.content_timemodified))

        except OSError:
            logging.debug(
                '[%d] Access time and modification time of the downloaded file could not be set', self.task_id
            )

    async def get_head_infos(self, dl_url: str) -> HeadInfo:
        """
        Do a Head request to collect some information about the URL
        @return: If download should be aborted then None; else HeadInfo
        """
        ssl_context = SslHelper.get_ssl_context(
            self.opts.global_opts.skip_cert_verify, self.opts.global_opts.allow_insecure_ssl
        )
        async with aiohttp.ClientSession(cookie_jar=self.get_cookie_jar(), raise_for_status=True) as session:
            try:
                async with session.request("HEAD", dl_url, headers=self.RQ_HEADER, ssl=ssl_context, timeout=20) as resp:
                    if resp.url != dl_url:
                        if resp.history and len(resp.history) > 0:
                            logging.debug('[%d] URL was %s time(s) redirected', self.task_id, len(resp.history))
                        else:
                            logging.debug('[%d] URL has changed after information retrieval', self.task_id)

                    guessed_file_name = posixpath.basename(resp.url.path)
                    if "Content-Disposition" in resp.headers.keys():
                        # Exp: Content-Disposition: attachment; filename="filename.jpg"
                        found_names = re.findall("filename=(.+)", resp.headers["Content-Disposition"])
                        if len(found_names) > 0:
                            guessed_file_name = unquote(found_names[0])

                    return HeadInfo(
                        # Exp: Content-Type: text/html; charset=utf-8
                        content_type=resp.headers.get('Content-Type', 'text/html').split(';')[0],
                        content_length=int(resp.headers.get('Content-Length', -1)),
                        # Exp: Last-Modified: Wed, 21 Oct 2015 07:28:00 GMT
                        last_modified=resp.headers.get('Last-Modified', None),
                        final_url=str(resp.url),
                        guessed_file_name=guessed_file_name,
                        host=resp.url.host,
                    )

            except aiohttp.InvalidURL:
                # don't download urls like 'mailto:name@provider.com'
                logging.debug(
                    '[%d] Download of the external file was canceled because the URL has an invalid format',
                    self.task_id,
                )
                return None
            except aiohttp.ClientResponseError as head_err:
                if head_err.status in [408, 409, 429]:
                    # 408 (timeout) or 409 (conflict) and 429 (too many requests)
                    logging.warning(
                        '[%d] Head request failed with status: %s %s', self.task_id, head_err.status, head_err.message
                    )
                    raise head_err from None

                logging.warning(
                    '[%d] Download of the external file was canceled because of HTTP error: %s %s',
                    self.task_id,
                    head_err.status,
                    head_err.message,
                )
                return None

            except (aiohttp.ClientError, OSError, ValueError, ContentRangeError) as head_err:
                logging.warning('[%d] Head request for external file failed with unexpected error', self.task_id)
                raise head_err from None

    async def download_using_yt_dlp(self, dl_url: str, infos: HeadInfo, delete_if_successful: bool):
        """
        @param delete_if_successful: Deletes the tmp file if download was successful
        @return: False if the page should be downloaded anyway; True if yt-dlp has processed the URL and we are done
        """
        # We try to limit the filename to < 250 chars
        if self.file.content_type == 'description-url':
            filename_template = '%(title).180B (%(id).32B).%(ext)s'
        else:
            filename_template = (
                PT.truncate_filename(self.filename, is_file=True, max_length=90) + ' - %(title).90B (%(id).32B).%(ext)s'
            )
        output_template = str(Path(self.destination) / filename_template)

        ydl_opts = {
            'logger': self.YtLogger(self),
            'progress_hooks': [self.yt_hook],
            'post_hooks': [self.yt_hook_after_move],
            'outtmpl': output_template,
            'nocheckcertificate': self.opts.global_opts.skip_cert_verify,
            'retries': 10,
            'fragment_retries': 10,
            'ignoreerrors': True,
            'addmetadata': True,
            'restrictfilenames': self.opts.restricted_filenames,
        }

        ydl_opts.update(self.opts.yt_dlp_options)

        if self.opts.cookies_text is not None:
            ydl_opts.update({'cookiefile': StringIO(self.opts.cookies_text)})

        ydl = yt_dlp.YoutubeDL(ydl_opts)
        add_additional_extractors(ydl)

        password_list = self.opts.video_passwords.get(infos.host, [None])
        if not isinstance(password_list, list):
            password_list = [password_list]
        if len(password_list) == 0:
            # Try at least once with no password
            password_list = [None]

        for password in password_list:
            if password is not None:
                # We allow to set videopassword via yt_dlp_options, so do not overwrite it with None
                ydl.params['videopassword'] = password

            # We restart yt-dlp, so we need to reset the return code
            self.status.yt_dlp_failed_with_error = False
            self.status.yt_dlp_used_generic_extractor = False
            ydl._download_retcode = 0  # pylint: disable=protected-access
            try:
                loop = asyncio.get_running_loop()
                ydl_result = await loop.run_in_executor(self.thread_pool, functools.partial(ydl.download, dl_url))
                # We set the saved_to path in yt_hook_after_move
                if ydl_result == 0:
                    if self.file.module_name == 'index_mod-page':
                        # We want to download legacy moodle pages
                        return False
                    # yt-dlp has an extractor for this URL so we do not want to download the URL extra
                    # only if yt-dlp used a generic extractor
                    return not self.status.yt_dlp_used_generic_extractor
            except Exception as yt_err:
                logging.error('[%d] yt-dlp failed! Error: %s', self.task_id, yt_err)
                self.status.yt_dlp_failed_with_error = True

        # If we want we could save ydl.cookiejar (Also the cookiejar of moodle-dl)

        if self.status.yt_dlp_failed_with_error and not self.opts.global_opts.ignore_ytdl_errors:
            if not delete_if_successful:
                PT.remove_file(self.file.saved_to)
            raise RuntimeError(
                'yt-dlp could not download the URL.'
                + ' You can ignore this error by running `moodle-dl --ignore-ytdl-errors` once.'
            )

        # We want to download the URL because yt-dlp has no extractor for it
        return False

    async def download_using_external_downloader(self, dl_url: str, external_dl_cmd: str, delete_if_successful: bool):
        cmd = external_dl_cmd.replace('%U', dl_url)
        logging.debug(
            '[%d] Run external downloader using the following command: `%s`',
            self.task_id,
            cmd,
        )
        external_dl_failed_with_error = False
        try:
            proc = await asyncio.create_subprocess_exec(
                shlex.split(cmd),
                cwd=str(self.destination),
                stderr=subprocess.PIPE,
                stdout=subprocess.PIPE,
            )

            for lines in await proc.stdout.readline():
                # line = line.decode('utf-8', 'replace')
                logging.debug('[%d] Ext-Dl: %s', self.task_id, lines.splitlines()[-1])

            _, stderr = await proc.communicate()

            if proc.returncode != 0:
                external_dl_failed_with_error = True
        except (subprocess.SubprocessError, ValueError, TypeError) as e:
            stderr = str(e)
            external_dl_failed_with_error = True

        if external_dl_failed_with_error:
            logging.error('[%d] External downloader error: %s', self.task_id, stderr)
            if not delete_if_successful:
                # cleanup the url-link file
                PT.remove_file(self.file.saved_to)
            raise RuntimeError('The external downloader could not download the URL')

        self.file.saved_to = str(Path(self.destination) / self.filename)

    async def external_download_url(self, add_token: bool, delete_if_successful: bool, needs_moodle_cookies: bool):
        """
        Use only for "external" shortcut/URL files.
        It tests whether a URL refers to a file, that is not an HTML web page then downloads it.
        Otherwise an attempt will be made to download it using yt-dlp.

        @param add_token: Adds the ws-token to the url
        @param delete_if_successful: Deletes the tmp file if download was successful
        @param needs_moodle_cookies: For this URL moodle cookies are required
        In case of an failure an exception will be raised
        """
        url_to_download = self.file.content_fileurl
        logging.debug('[%d] Try to download external file %s', self.task_id, url_to_download)

        if add_token:
            url_to_download = self.add_token_to_url(url_to_download)

        if delete_if_successful:
            # If temporary file is not needed delete it as soon as possible
            PT.remove_file(self.file.saved_to)

        if needs_moodle_cookies and self.opts.cookies_text is None:
            # Without Moodle cookies we should not continue
            # We assume that there are Moodle cookies in the cookie file, if one exists.
            # TODO: Perhaps explicitly check if Moodle cookies are set
            raise ValueError(
                'Moodle cookies are missing. Set a private token so that moodle-dl can obtain moodle cookies'
            )

        infos = await self.get_head_infos(url_to_download)
        if infos is None:
            # Head request failed but we declare it as success (because URL is broken)
            return

        external_dl_cmd = self.opts.external_file_downloaders.get(infos.host, "")
        if infos.is_html and external_dl_cmd != "":
            await self.download_using_external_downloader(
                dl_url=url_to_download,
                external_dl_cmd=external_dl_cmd,
                delete_if_successful=delete_if_successful,
            )
            return
        if infos.is_html and not self.is_blocked_for_yt_dlp(url_to_download):
            yt_dlp_processed = await self.download_using_yt_dlp(
                dl_url=url_to_download,
                infos=infos,
                delete_if_successful=delete_if_successful,
            )
            if yt_dlp_processed:
                return

        logging.debug('[%d] Downloading URL directly', self.task_id)

        # Generate file name for external file
        new_name, new_extension = os.path.splitext(infos.guessed_file_name)
        if new_extension == '' or infos.is_html:
            new_extension = '.html'

        if self.file.content_type == 'description-url' and new_name != '':
            self.filename = new_name + new_extension

        _old_name, old_extension = os.path.splitext(self.filename)
        if old_extension != new_extension:
            self.filename = self.filename + new_extension

        self.set_path(True)

        await self.download_url(url_to_download, self.file.saved_to)

    def is_filtered_external_domain(self):
        """
        Filter external linked files.
        Check if the domain of the download link is on the blacklist or is not on the whitelist.

        @return: True if the domain is filtered.
        """

        domain = urlparse.urlparse(self.file.content_fileurl).hostname

        if domain is None or domain == '':
            return True

        in_blacklist = False
        for entry in self.opts.download_domains_blacklist:
            if domain == entry or domain.endswith('.' + entry):
                in_blacklist = True
                break

        in_whitelist = len(self.opts.download_domains_whitelist) == 0
        for entry in self.opts.download_domains_whitelist:
            if domain == entry or domain.endswith('.' + entry):
                in_whitelist = True
                break

        return not in_whitelist or in_blacklist

    async def create_shortcut(self):
        "Create a Shortcut to a URL"
        logging.debug('[%d] Creating a shortcut', self.task_id)
        PT.remove_file(self.file.saved_to)
        for link_type, should_write in self.opts.write_links.items():
            if should_write:
                self.set_path(True, link_type)
                async with aiofiles.open(
                    self.file.saved_to, 'w+', encoding='utf-8', newline='\r\n' if link_type == 'url' else '\n'
                ) as shortcut:
                    template_vars = {'url': self.file.content_fileurl}
                    if link_type == 'desktop':
                        template_vars['filename'] = self.file.saved_to[: -(len(link_type) + 1)]
                    await shortcut.write(LINK_TEMPLATES[link_type] % template_vars)

    def set_path(self, ignore_attributes: bool = False, force_file_extension=None):
        """Set the path where a file should be created. The file type is used to set the needed file extension.
        An empty target file is created which may need to be cleaned up.

        @param ignore_attributes: If the file attributes should be ignored.
        """
        if self.file.content_type == 'description' and not ignore_attributes:
            self.file.saved_to = str(Path(self.destination) / (self.filename + '.md'))

        elif self.file.content_type == 'html' and not ignore_attributes:
            self.file.saved_to = str(Path(self.destination) / (self.filename + '.html'))

        elif force_file_extension is not None:
            self.file.saved_to = str(Path(self.destination) / (self.filename + f'.{force_file_extension}'))

        else:  # normal path
            self.file.saved_to = str(Path(self.destination) / self.filename)

        self.file.saved_to = self.create_target_file(self.file.saved_to)

    async def create_description(self):
        "Create a description file"
        logging.debug('[%d] Creating a description file', self.task_id)

        md_content = ''
        if self.file.text_content is not None:
            h2t_handler = html2text.HTML2Text()
            md_content = h2t_handler.handle(self.file.text_content).strip()
            # we could run html.unescape() over to_save, but this could destroy the md file

        if md_content == '':
            logging.debug('[%d] Remove target file because description file would be empty', self.task_id)
            os.remove(self.file.saved_to)
            return

        async with aiofiles.open(self.file.saved_to, 'w+', encoding='utf-8') as md_file:
            await md_file.write(md_content)

    async def create_html_file(self):
        "Create a HTML file"
        logging.debug('[%d] Creating a html file', self.task_id)

        html_content = ''
        if self.file.html_content is not None:
            html_content = self.file.html_content

        if html_content == '':
            logging.debug('[%d] Remove target file because html file would be empty', self.task_id)
            os.remove(self.file.saved_to)
            return

        async with aiofiles.open(self.file.saved_to, 'w+', encoding='utf-8') as html_file:
            await html_file.write(html_content)

    def move_old_file(self) -> bool:
        """
        Try to move the old file to the new location.
        @return: True if successful. Else the file needs to be re-downloaded.
        """

        if self.file.old_file is None:
            return False

        old_path = self.file.old_file.saved_to
        if not os.path.exists(old_path):
            return False

        logging.debug('[%d] Moving old file "%s" to new target location', self.task_id, old_path)
        try:
            # On Windows, the temporary file must be deleted first.
            os.remove(self.file.saved_to)
            shutil.move(old_path, self.file.saved_to)
            return True
        except OSError as e:
            logging.warning('[%d] Moving the old file %s failed unexpectedly!  Error: %s', self.task_id, old_path, e)
        return False

    async def create_data_url_file(self):
        url_to_download = self.file.content_fileurl
        logging.debug('[%d] Creating a Data-URL file', self.task_id)
        PT.remove_file(self.file.saved_to)
        self.set_path(True)
        with urllib.request.urlopen(url_to_download) as response:
            data = response.read()

        async with aiofiles.open(self.file.saved_to, "wb") as target_file:
            await target_file.write(data)

    async def run(self):
        if self.status.state != TaskState.INIT:
            logging.debug('[%d] Task was already started', self.task_id)
            return
        self.status.state = TaskState.STARTED

        success = await self.real_run()

        if success:
            self.set_utime()
            self.file.time_stamp = int(time.time())

    async def real_run(self) -> bool:
        try:
            logging.debug('[%d] Starting Task: %s', self.task_id, self)
            PT.make_dirs(self.destination)

            # If file was modified try rename the old file, before create new one
            if self.file.modified:
                self.rename_old_file()

            # Create an empty destination file
            self.set_path()
            logging.debug('[%d] Starting downloading of: %s', self.task_id, self.file.saved_to)

            # Try to move the old file if it still exists
            if self.file.moved:
                if self.move_old_file():
                    return True

            if self.file.content_type == 'description':
                # Create a description file instead of downloading it
                await self.create_description()

            elif self.file.content_type == 'html':
                # Create a HTML file instead of downloading it
                await self.create_html_file()

            elif self.file.module_modname.startswith('index_mod'):
                await self.external_download_url(add_token=True, delete_if_successful=True, needs_moodle_cookies=False)

            elif self.file.module_modname.startswith('cookie_mod'):
                await self.external_download_url(add_token=False, delete_if_successful=True, needs_moodle_cookies=True)

            elif self.file.module_modname.startswith('url') and not self.file.content_fileurl.startswith('data:'):
                # Create a shortcut and maybe downloading it
                await self.create_shortcut()
                if self.opts.download_linked_files and not self.is_filtered_external_domain():
                    await self.external_download_url(
                        add_token=False, delete_if_successful=False, needs_moodle_cookies=False
                    )

            elif self.file.content_fileurl.startswith('data:'):
                await self.create_data_url_file()

            else:
                url_to_download = self.file.content_fileurl
                logging.debug('[%d] Downloading %s', self.task_id, url_to_download)
                url_to_download = self.add_token_to_url(self.file.content_fileurl)
                await self.download_url(url_to_download, self.file.saved_to)

            logging.debug('[%d] Download finished', self.task_id)
            self.report_success()
            return True
        except Exception as dl_err:
            self.status.error = dl_err

            logging.error('[%d] %r', self.task_id, dl_err)
            logging.error('[%d] Error while trying to download file: %s', self.task_id, dl_err)

            if os.path.isfile(self.file.saved_to):
                file_size = 0
                try:
                    file_size = os.path.getsize(self.file.saved_to)
                except OSError:
                    pass
                logging.debug(
                    '[%d] file size: %d; downloaded: %d',
                    self.task_id,
                    file_size,
                    self.status.bytes_downloaded,
                )

            logging.debug('[%d] Traceback:\n%s', self.task_id, traceback.format_exc())

            # TODO: Do this in the error handlers of download functions
            # TODO: See download_url; remove only if not recoverable
            PT.remove_file(self.file.saved_to)
            self.report_received_bytes(-self.status.bytes_downloaded)
            self.report_failure()

        return False

    def get_cookie_jar(self) -> aiohttp.CookieJar:
        # TODO: Since we currently do not modify the cookieJar we could just use a deep copied instance.
        if self.opts.cookies_text is not None:
            cookie_jar = MoodleDLCookieJar(StringIO(self.opts.cookies_text))
            cookie_jar.load(ignore_discard=True, ignore_expires=True)
            return convert_to_aiohttp_cookie_jar(cookie_jar)
        return None

    async def check_range_download_opt(self, url, session):
        try:
            headers = self.RQ_HEADER.copy()
            headers['Range'] = 'bytes=0-4'
            resp = await session.request("GET", url, headers=headers)
            return resp.headers.get('Content-Range') is not None and resp.status == 206
        except Exception as err:
            logging.debug("Failed to check if download can be continued on fail: %s", err)
        return False

    def report_success(self):
        self.status.state = TaskState.FINISHED
        self.callback(DlEvent.FINISHED, self)

    def report_failure(self):
        self.status.state = TaskState.FAILED
        self.callback(DlEvent.FAILED, self)

    def report_received_bytes(self, bytes_received: int):
        self.status.bytes_downloaded += bytes_received
        self.callback(DlEvent.RECEIVED, self, bytes_received=bytes_received)

    def report_content_length(self, content_length: int, save_in_status: bool = True):
        if content_length is not None and content_length != 0:
            if self.file.content_filesize is None or self.file.content_filesize <= 0:
                if save_in_status:
                    self.status.external_total_size = content_length
                self.callback(DlEvent.TOTAL_SIZE, self, content_length=content_length)

    async def download_url(self, dl_url: str, dest_path: str, timeout: int = None):
        total_bytes_received = 0
        done_tries = 0
        can_continue_on_fail = False
        file_obj = None
        headers = self.RQ_HEADER.copy()
        ssl_context = SslHelper.get_ssl_context(
            self.opts.global_opts.skip_cert_verify, self.opts.global_opts.allow_insecure_ssl
        )
        with Timer() as watch:
            async with aiohttp.ClientSession(cookie_jar=self.get_cookie_jar(), raise_for_status=True) as session:
                while done_tries < self.MAX_DL_RETRIES:
                    try:
                        if done_tries > 0:
                            logging.debug(
                                '[%d] Start downloading (Try %d of %d)',
                                self.task_id,
                                done_tries + 1,
                                self.MAX_DL_RETRIES,
                            )

                        if done_tries > 0 and can_continue_on_fail:
                            headers['Range'] = f'bytes={total_bytes_received}-'
                        elif not can_continue_on_fail and 'Range' in headers:
                            del headers['Range']

                        async with session.request(
                            "GET", dl_url, headers=headers, ssl=ssl_context, timeout=timeout
                        ) as resp:
                            content_length = int(resp.headers.get("Content-Length", 0))
                            self.report_content_length(content_length)
                            content_range = resp.headers.get("Content-Range")  # Exp: bytes 200-1000/67589

                            if resp.status not in [200, 206]:
                                logging.debug('[%d] Warning got status %s', self.task_id, resp.status)

                            if done_tries > 0 and can_continue_on_fail and not content_range and resp.status != 206:
                                raise ContentRangeError(
                                    f"[{self.task_id}] Server did not response with requested range data"
                                )

                            file_obj = file_obj or await aiofiles.open(dest_path, "wb")
                            async for chunk in resp.content.iter_chunked(self.CHUNK_SIZE):
                                bytes_received = len(chunk)
                                total_bytes_received += bytes_received
                                self.report_received_bytes(bytes_received)
                                await file_obj.write(chunk)

                        if file_obj is not None and not file_obj.closed:
                            await file_obj.close()

                        if content_length >= 0 and total_bytes_received < content_length:
                            raise ContentTooShortError(
                                f'[{self.task_id}] Download incomplete: Got only {format_bytes(total_bytes_received)}'
                                + f' out of {format_bytes(content_length)} bytes',
                                dest_path,
                            )

                        logging.debug('[%d] Successfully downloaded %s', self.task_id, dest_path)
                        break

                    except (aiohttp.ClientError, OSError, ValueError, ContentRangeError) as err:
                        if done_tries == 0:
                            can_continue_on_fail = await self.check_range_download_opt(dl_url, session)

                        done_tries += 1
                        if (
                            (not can_continue_on_fail and total_bytes_received > 0)
                            or isinstance(err, ContentRangeError)
                            or (done_tries >= self.MAX_DL_RETRIES)
                        ):
                            can_continue_on_fail = False
                            # Clean up failed file because we can not recover
                            if file_obj is not None and not file_obj.closed:
                                await file_obj.close()
                            file_obj = None

                            # TODO: If download can be continued and size > 0,
                            #  remember that the file started downloading, and continue downloading
                            #  on next run instead of deleting it.
                            PT.remove_file(dest_path)
                            self.report_received_bytes(-total_bytes_received)
                            total_bytes_received = 0

                        if isinstance(err, aiohttp.ClientResponseError):
                            if err.status not in [408, 409, 429]:  # pylint: disable=no-member
                                # 408 (timeout) or 409 (conflict) and 429 (too many requests)
                                logging.warning(
                                    '[%d] Download failed with status: %s %s', self.task_id, err.status, err.message
                                )
                                raise err from None

                        if done_tries < self.MAX_DL_RETRIES:
                            logging.debug('[%d] Download error occurred: %s', self.task_id, err)
                            await asyncio.sleep(1)
                            continue

                        # No more tries
                        raise err from None
        logging.debug(
            '[%d] Download of %s finished in %s',
            self.task_id,
            format_bytes(total_bytes_received),
            format_seconds(watch.duration),
        )

    def __str__(self):
        return 'Task (%(task_id)s, %(file)s, %(course)s, %(status)s)' % {
            'task_id': self.task_id,
            'file': self.file,
            'course': self.course,
            'status': self.status,
        }


class ContentRangeError(ConnectionError):
    pass



import logging
from typing import Dict, List

from moodle_dl.config import ConfigHelper
from moodle_dl.moodle.mods import MoodleMod
from moodle_dl.moodle.request_helper import RequestRejectedError
from moodle_dl.types import Course, File
from moodle_dl.utils import PathTools as PT


class AssignMod(MoodleMod):
    MOD_NAME = 'assign'
    MOD_PLURAL_NAME = 'assignments'
    MOD_MIN_VERSION = 2012120300  # 2.4

    @classmethod
    def download_condition(cls, config: ConfigHelper, file: File) -> bool:
        # TODO: Add condition for assignments not only submissions
        return config.get_download_submissions() or (not (file.module_modname.endswith(cls.MOD_NAME) and file.deleted))

    async def real_fetch_mod_entries(
        self, courses: List[Course], core_contents: Dict[int, List[Dict]]
    ) -> Dict[int, Dict[int, Dict]]:
        assign_courses = (
            await self.client.async_post('mod_assign_get_assignments', self.get_data_for_mod_entries_endpoint(courses))
        ).get('courses', [])

        result = {}
        for assign_course in assign_courses:
            course_id = assign_course.get('id', 0)
            result[course_id] = self.extract_assign_modules(assign_course.get('assignments', []))

        await self.add_submissions(result)
        await self.add_foreign_submissions(result)
        return result

    def extract_assign_modules(self, assignments: List[Dict]) -> Dict[int, Dict]:
        result = {}
        for assign in assignments:
            assign_files = assign.get('introfiles', [])
            assign_files += assign.get('introattachments', [])

            self.set_props_of_files(assign_files, type='assign_file')

            assign_intro = assign.get('intro', '')
            if assign_intro != '':
                assign_files.append(
                    {
                        'filename': 'Assignment intro',
                        'filepath': '/',
                        'description': assign_intro,
                        'type': 'description',
                    }
                )

            result[assign.get('cmid', 0)] = {
                'id': assign.get('id', 0),
                'name': assign.get('name', ''),
                'timemodified': assign.get('timemodified', 0),
                'files': assign_files,
            }
        return result

    async def add_submissions(self, assignments: Dict[int, Dict[int, Dict]]):
        """
        Fetches for the assignments list additionally the submissions
        @param assignments: Dictionary of all assignments, indexed by courses, then module id
        """
        if not self.config.get_download_submissions():
            return

        if self.version < 2016052300:  # 3.1
            return

        await self.run_async_load_function_on_mod_entries(assignments, self.load_submissions)

    async def add_foreign_submissions(self, assignments: Dict[int, Dict[int, Dict]]):
        """
        Fetches for the assignments list additionally the submissions of other students
        @param assignments: Dictionary of all assignments, indexed by courses, then module id
        """
        if not self.config.get_download_submissions():
            return

        if self.version < 2013051400:  # 2.5
            return

        # get submissions of all students for all assignments (only teachers can see that)
        indexed_assignment_ids = self.get_indexed_ids_of_mod_instances(assignments)
        if len(indexed_assignment_ids) <= 0:
            return

        assignments_with_all_submissions = (
            await self.client.async_post('mod_assign_get_submissions', {'assignmentids': indexed_assignment_ids})
        ).get('assignments', [])

        if len(assignments_with_all_submissions) <= 0:
            return

        for course_id, modules in assignments.items():
            found_assignment_in_course = False
            for assignment in assignments_with_all_submissions:
                for _module_id, module in modules.items():
                    if assignment['assignmentid'] == module['id']:
                        found_assignment_in_course = True
                        break
                if found_assignment_in_course:
                    break
            if not found_assignment_in_course:
                continue
            # TODO: Extract the API call to get enrolled users, if we need the information also in another mod
            try:
                course_users = await self.client.async_post('core_enrol_get_enrolled_users', {'courseid': course_id})
            except RequestRejectedError:
                logging.debug("No access rights for enrolled users list of course %d", course_id)
                return

            for assignment in assignments_with_all_submissions:
                found_module = None
                for _module_id, module in modules.items():
                    if assignment['assignmentid'] == module['id']:
                        found_module = module
                        break
                if found_module is None:
                    continue

                for submission in assignment.get('submissions', []):
                    user_id = submission.get('userid', 0)
                    group_id = submission.get('groupid', 0)
                    subfolder = None
                    if user_id == 0:
                        # Its a group submission
                        found_users = []
                        group_name = None
                        for user in course_users:
                            for group in user.get('groups', []):
                                if group.get('id', 0) == group_id:
                                    found_users.append(user)
                                    if group_name is None:
                                        group_name = group.get('name')
                                    break
                        if len(found_users) == 0:
                            # should not happen
                            continue
                        all_usernames = ' & '.join(
                            (f"{user.get('fullname')} ({user.get('idnumber') or user.get('id', 0)})")
                            for user in found_users
                        )
                        subfolder = PT.to_valid_name(
                            f"{group_name or 'Unnamed group'} ({group_id}): {all_usernames}", is_file=False
                        )
                    else:
                        # Its a user submission
                        found_user = None
                        for user in course_users:
                            if user.get('id', 0) == user_id:
                                found_user = user
                                break
                        if found_user is None:
                            # should not happen
                            continue
                        subfolder = PT.to_valid_name(
                            f"{found_user.get('fullname')} ({found_user.get('idnumber') or found_user.get('id', 0)})",
                            is_file=False,
                        )
                    found_module['files'] += self._get_files_of_plugins(submission, f'/all_submissions/{subfolder}/')

    async def load_submissions(self, assign: Dict):
        "Fetches for a given assign module the submissions"
        data = {'userid': self.user_id, 'assignid': assign.get('id', 0)}
        submission = await self.client.async_post('mod_assign_get_submission_status', data)
        assign['files'] += self._get_files_of_submission(submission)

    def _get_files_of_submission(self, submission: Dict) -> List[Dict]:
        result = []
        # get own submission
        last_attempt = submission.get('lastattempt', {})
        last_submission = last_attempt.get('submission', {})
        last_team_submission = last_attempt.get('teamsubmission', {})
        # We could also extract previous attempts, but for now we are only interested in last attempt
        # Multiple attempts on assignments are very raw and therefore not implemented yet

        # get teachers feedback
        feedback = submission.get('feedback', {})

        base_file_path = '/submissions/'
        result += self._get_files_of_plugins(last_submission, base_file_path)
        result += self._get_files_of_plugins(last_team_submission, base_file_path)
        result += self._get_files_of_plugins(feedback, base_file_path)
        result += self._get_grade_of_feedback(feedback, base_file_path)

        return result

    def _get_grade_of_feedback(self, feedback: Dict, base_file_path: str) -> List[Dict]:
        grade_for_display = feedback.get('gradefordisplay')
        graded_date = feedback.get('gradeddate')
        if graded_date is None or grade_for_display is None:
            return []

        return [
            {
                'filename': 'grade',
                'filepath': base_file_path,
                'timemodified': graded_date,
                'description': grade_for_display,
                'type': 'description',
            }
        ]

    def _get_files_of_plugins(self, obj: Dict, base_file_path: str) -> List[Dict]:
        result = []
        plugins = obj.get('plugins', [])

        for plugin in plugins:
            # We could use the plugin name in the file structure, but it is most of the time unnecessary information
            for file_area in plugin.get('fileareas', []):
                files = file_area.get('files', [])
                self.set_props_of_files(files, type='submission_file')
                self.set_base_file_path_of_files(files, base_file_path)
                result += files

            for editor_field in plugin.get('editorfields', []):
                filename = editor_field.get('description', '')
                description = editor_field.get('text', '')
                if filename != '' and description != '':
                    result.append(
                        {
                            'filename': filename,
                            'description': description,
                            'type': 'description',
                            'filepath': base_file_path,
                        }
                    )

        return result



import asyncio
import logging
import math
from abc import ABCMeta, abstractmethod
from typing import Dict, List

from moodle_dl.config import ConfigHelper
from moodle_dl.moodle.request_helper import RequestHelper
from moodle_dl.types import Course, File
from moodle_dl.utils import PathTools as PT
from moodle_dl.utils import get_nested, run_with_final_message


class MoodleMod(metaclass=ABCMeta):
    """
    Common class for a Moodle module endpoint
    """

    MOD_NAME = None
    MOD_PLURAL_NAME = None
    MOD_MIN_VERSION = None

    def __init__(
        self,
        request_helper: RequestHelper,
        moodle_version: int,
        user_id: int,
        last_timestamps: Dict[str, Dict[int, int]],
        config: ConfigHelper,
    ):
        """
        @param last_timestamps: A Dict per mod of timestamps per course module id,
                                prevents downloading older content of a corse module
        """

        self.client = request_helper
        self.version = moodle_version
        self.user_id = user_id
        self.last_timestamps = last_timestamps
        self.config = config

    @classmethod
    @abstractmethod
    def download_condition(cls, config: ConfigHelper, file: File) -> bool:
        """
        Return True if moodle-dl is configured to downloaded the given file
        This condition is applied after comparing the current status with the local database
        """
        # TODO: Make module download conditions more granular and more generally
        # (do not only filter "deleted" mod files but all?)
        pass

    async def fetch_mod_entries(
        self, courses: List[Course], core_contents: Dict[int, List[Dict]]
    ) -> Dict[int, Dict[int, Dict]]:
        if self.version < self.MOD_MIN_VERSION:
            return {}

        result = await self.real_fetch_mod_entries(courses, core_contents)
        logging.info('Loaded all %s', self.MOD_PLURAL_NAME)
        return result

    def get_indexed_ids_of_mod_instances(self, mod_instances: Dict[int, Dict[int, Dict]]):
        """
        @param mod_instances: Dictionary of all mod instanced, indexed by courses, then module id
        """
        mod_instances_ids = {}
        idx = 0
        for _, modules in mod_instances.items():
            for _, mod in modules.items():
                mod_instances_ids[str(idx)] = mod['id']
                idx += 1
        return mod_instances_ids

    def get_data_for_mod_entries_endpoint(self, courses: List[Course]):
        # Create a dictionary with all the courses we want to request
        course_ids = {}
        for idx, course in enumerate(courses):
            course_ids[str(idx)] = course.id
        return {'courseids': course_ids}

    @abstractmethod
    async def real_fetch_mod_entries(
        self, courses: List[Course], core_contents: Dict[int, List[Dict]]
    ) -> Dict[int, Dict[int, Dict]]:
        """
        Fetches the mod entries for all courses
        @return: Dictionary of all course modules of that mod type, indexed by course id, then course module id
        """
        pass

    @staticmethod
    def set_props_of_file(file_dict: Dict, **props):
        for prop, value in props.items():
            if (
                file_dict.get(prop) is not None
                and file_dict[prop] != ''
                and file_dict[prop] != value
                and (prop != 'filepath' or file_dict[prop] != '/')
            ):
                logging.debug('Overwriting %r with %r of file %s', prop, value, file_dict)
            file_dict[prop] = value

    @classmethod
    def set_props_of_files(cls, files: List[Dict], **props):
        for file_dict in files:
            cls.set_props_of_file(file_dict, **props)

    @staticmethod
    def set_base_file_path_of_file(file_dict: Dict, base_file_path: str):
        old_file_path = file_dict.get('filepath')
        if old_file_path is not None and old_file_path != '/':
            file_dict['filepath'] = PT.make_path(base_file_path, old_file_path.strip('/'))
        else:
            file_dict['filepath'] = base_file_path

    @classmethod
    def set_base_file_path_of_files(cls, files: List[Dict], base_file_path: str):
        for file_dict in files:
            cls.set_base_file_path_of_file(file_dict, base_file_path)

    @classmethod
    async def run_async_load_function_on_mod_entries(cls, entries: Dict[int, Dict[int, Dict]], load_function):
        """
        Runs a load function on every module in a given entries list
        @param entries: Dictionary of all module entries, indexed by courses, then module id
        """
        ctr = 0
        total_entries = 0
        for _, entries_in_course in entries.items():
            total_entries += len(entries_in_course)

        if total_entries == 0:
            return
        ctr_digits = int(math.log10(total_entries)) + 1

        async_features = []
        for course_id, entries_in_course in entries.items():
            for module_id, entry in entries_in_course.items():
                ctr += 1

                # Example: [ 5/16] Loaded assign 123 in course 456 "Assignment name"
                loaded_message = (
                    f'[%(ctr){ctr_digits}d/%(total){ctr_digits}d] Loaded %(mod_name)s %(module_id)d'
                    + ' in course %(course_id)d "%(module_name)s"'
                )

                async_features.append(
                    run_with_final_message(
                        load_function,
                        entry,
                        loaded_message,
                        {
                            'ctr': ctr,
                            'total': total_entries,
                            'mod_name': cls.MOD_NAME,
                            'module_id': module_id,
                            'course_id': course_id,
                            'module_name': entry.get('name', ''),
                        },
                    )
                )

        await asyncio.gather(*async_features)

    @classmethod
    async def run_async_collect_function_on_list(
        cls,
        entries: List[Dict],
        collect_function,
        collect_kind: str,
        format_mapping: Dict,
    ) -> List[Dict]:
        "Runs a collect function on every entry in a given entries list"
        total_entries = len(entries)
        if total_entries == 0:
            return []
        ctr_digits = int(math.log10(total_entries)) + 1

        async_features = []
        for ctr, entry in enumerate(entries):
            # Example: [ 5/16] Loaded forum discussion 123 "Good discussion"
            loaded_message = (
                f'[%(ctr){ctr_digits}d/%(total){ctr_digits}d] Loaded %(mod_name)s %(collect_kind)s'
                + ' %(collect_id)d "%(collect_name)s"'
            )

            async_features.append(
                run_with_final_message(
                    collect_function,
                    entry,
                    loaded_message,
                    {
                        'ctr': ctr + 1,
                        'total': total_entries,
                        'mod_name': cls.MOD_NAME,
                        'collect_kind': collect_kind,
                        'collect_id': get_nested(entry, format_mapping['collect_id'], 0),
                        'collect_name': get_nested(entry, format_mapping['collect_name'], ''),
                    },
                )
            )

        result = []
        for feature_result in await asyncio.gather(*async_features):
            if isinstance(feature_result, list):
                result.extend(feature_result)
            elif feature_result is not None:
                result.append(feature_result)
        return result

    @staticmethod
    def get_module_in_core_contents(course_id: int, module_id: int, core_contents: Dict[int, List[Dict]]) -> Dict:
        course_section = core_contents.get(course_id, [])
        for section in course_section:
            for module in section.get('modules', []):
                if module.get('id', 0) == module_id:
                    return module
        return {}

    @staticmethod
    def add_module(result: Dict, course_id: int, module_id: int, module: Dict):
        if course_id not in result:
            result[course_id] = {}
        if module_id in result[course_id]:
            logging.warning('Got duplicated module %s in course %s', module_id, course_id)
        result[course_id][module_id] = module



import asyncio
import logging
import time
from concurrent.futures import ThreadPoolExecutor
from typing import List

from moodle_dl.config import ConfigHelper
from moodle_dl.database import StateRecorder
from moodle_dl.downloader.task import Task
from moodle_dl.types import Course, DlEvent, DownloadStatus, MoodleDlOpts, TaskState
from moodle_dl.utils import calc_speed, format_bytes, format_speed


class DownloadService:
    "Manages jobs to download, delete or create files of courses"

    def __init__(self, courses: List[Course], config: ConfigHelper, opts: MoodleDlOpts, database: StateRecorder):
        self.courses = courses
        self.config = config
        self.opts = opts
        self.database = database

        self.status = DownloadStatus()
        self.all_tasks = self.gen_all_tasks()

    def gen_all_tasks(self) -> List:
        # Set custom chunk size
        Task.CHUNK_SIZE = self.opts.download_chunk_size
        dl_options = self.config.get_download_options(self.opts)
        thread_pool = ThreadPoolExecutor(max_workers=self.opts.max_parallel_yt_dlp)
        all_tasks = []
        for course in self.courses:
            for course_file in course.files:
                if course_file.deleted is False:
                    all_tasks.append(
                        Task(
                            task_id=self.status.files_to_download,
                            file=course_file,
                            course=course,
                            options=dl_options,
                            thread_pool=thread_pool,
                            callback=self.status_callback,
                        )
                    )
                    self.status.bytes_to_download += course_file.content_filesize
                    self.status.files_to_download += 1
        if self.status.files_to_download > 0:
            logging.info('Download queue contains %d tasks', self.status.files_to_download)
        else:
            logging.debug('Download queue is empty')
        return all_tasks

    def status_callback(self, event: DlEvent, task: Task, **extra_args):
        self.status.lock.acquire()
        if event == DlEvent.RECEIVED:
            self.status.bytes_downloaded += extra_args['bytes_received']
        elif event == DlEvent.FAILED:
            self.status.files_failed += 1
        elif event == DlEvent.FINISHED:
            self.database.save_file(task.file, task.course.id, task.course.fullname)
            self.status.files_downloaded += 1
        elif event == DlEvent.TOTAL_SIZE:
            self.status.bytes_to_download += extra_args['content_length']
        elif event == DlEvent.TOTAL_SIZE_UPDATE:
            self.status.bytes_to_download += extra_args['content_length_diff']
        self.status.lock.release()

    def run(self):
        asyncio.run(self.real_run())

    async def real_run(self):
        "Starts all tasks and issues status messages at regular intervals"

        # delete files, that should be deleted
        self.database.batch_delete_files(self.courses)

        if len(self.all_tasks) <= 0:
            return

        # run all other tasks
        status_logger_task = asyncio.create_task(self.log_download_status())

        dl_tasks = set()
        for task in self.all_tasks:
            if len(dl_tasks) >= self.opts.max_parallel_downloads:
                # Wait for some download to finish before adding a new one
                _done, dl_tasks = await asyncio.wait(dl_tasks, return_when=asyncio.FIRST_COMPLETED)
            dl_tasks.add(asyncio.create_task(task.run()))

        # Wait for the remaining downloads to finish
        await asyncio.wait(dl_tasks)
        status_logger_task.cancel()

    async def log_download_status(self):
        last_bytes_downloaded = 0
        last_status_timestamp = time.time()
        while True:
            # Print every 2 sec the current status
            await asyncio.sleep(2)

            percentage = None
            if self.status.bytes_to_download != 0:
                percentage = int(self.status.bytes_downloaded * 100 / self.status.bytes_to_download)
                if percentage > 100 or percentage < 0:
                    percentage = None
            if percentage is None:
                percentage = ' NA%'
            else:
                percentage = f'{percentage:3}%'

            speed = calc_speed(last_status_timestamp, time.time(), self.status.bytes_downloaded - last_bytes_downloaded)
            last_status_timestamp = time.time()
            last_bytes_downloaded = self.status.bytes_downloaded

            message_line = (
                f'Total: {percentage}'
                + f' {format_bytes(self.status.bytes_downloaded):>5} / {format_bytes(self.status.bytes_to_download):<5}'
                + f' | Done: {(self.status.files_downloaded + self.status.files_failed):>5}'
                + f' / {self.status.files_to_download:<5}'
                + f' | Speed: {format_speed(speed)}'
            )
            if self.status.files_failed > 0:
                message_line += f' | Failed: {self.status.files_failed}'

            logging.info(message_line)

    def get_failed_tasks(self) -> List[Task]:
        "Return a list of failed downloads."
        result = []
        for task in self.all_tasks:
            if task.status.state == TaskState.FAILED:
                result.append(task)
        return result



from os import path

from setuptools import find_packages, setup

# Get the version from moodle_dl/version.py without importing the package
exec(compile(open('moodle_dl/version.py').read(), 'moodle_dl/version.py', 'exec'))


def readme():
    this_directory = path.abspath(path.dirname(__file__))
    with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
        return f.read()


setup(
    name='moodle-dl',
    version=__version__,
    description='Moodle-DL downloads course content fast from Moodle (eg. lecture pdfs)',
    long_description=readme(),
    long_description_content_type='text/markdown',
    url='https://github.com/C0D3D3V/Moodle-DL',
    project_urls={
        'Documentation': 'https://github.com/C0D3D3V/Moodle-DL#readme',
        'Wiki': 'https://github.com/C0D3D3V/Moodle-DL/wiki',
        'Source': 'https://github.com/C0D3D3V/Moodle-DL',
        'Tracker': 'https://github.com/C0D3D3V/Moodle-DL/issues',
    },
    author='C0D3D3V',
    author_email='moodle-dl@ist-ein-knaller.de',
    license='GPL-3.0',
    packages=find_packages(),
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'moodle-dl = moodle_dl.main:main',
        ],
    },
    python_requires='>=3.7',
    install_requires=[
        'aiodns>=3.0.0',
        'aiofiles>=22.1.0',
        'aiohttp>=3.8.4',
        'certifi>=2020.4.5.2',
        'colorama>=0.4.6',
        'colorlog>=6.7.0',
        'html2text>=2020.1.16',
        'readchar>=4.0.3',
        'requests>=2.24.0',
        'sentry_sdk>=0.13.5',
        'xmpppy>=0.7.1',
        'yt_dlp>=2024.3.10',
    ],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: End Users/Desktop',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Education',
        'Topic :: Internet :: WWW/HTTP :: Indexing/Search',
        'Topic :: Multimedia :: Video',
        'Topic :: Multimedia :: Sound/Audio',
        'Topic :: Utilities',
    ],
    zip_safe=False,
)



from datetime import datetime
from typing import Dict, List

from moodle_dl.config import ConfigHelper
from moodle_dl.moodle.mods import MoodleMod
from moodle_dl.moodle.moodle_constants import (
    course_events_module_id,
    course_events_section_id,
    moodle_event_footer,
    moodle_event_header,
)
from moodle_dl.types import Course, File
from moodle_dl.utils import PathTools as PT

# TODO: should we use locale.setlocale(locale.LC_TIME, "") to set localized version of Date format
# or should we use https://babel.pocoo.org/en/latest/dates.html
# babel.dates.format_datetime(datetime_obj) is enough to use local setting, but we could also use user language settings of moodle,
# by defining locale='user_lang' as extra argument
# We can also use babel.dates.format_timedelta(time_delta) to print the time delta into the date file


class CalendarMod(MoodleMod):
    MOD_NAME = 'calendar'
    MOD_PLURAL_NAME = 'events'
    MOD_MIN_VERSION = 2013051400  # 2.5

    @classmethod
    def download_condition(cls, config: ConfigHelper, file: File) -> bool:
        return config.get_download_calendars() or (not (file.module_modname.endswith(cls.MOD_NAME) and file.deleted))

    async def real_fetch_mod_entries(
        self, courses: List[Course], core_contents: Dict[int, List[Dict]]
    ) -> Dict[int, Dict[int, Dict]]:
        result = {}
        if not self.config.get_download_calendars():
            return result

        last_timestamp = self.last_timestamps.get(self.MOD_NAME, {}).get(course_events_module_id, 0)
        calendar_req_data = {
            'options': {'timestart': last_timestamp, 'userevents': 0},
            'events': self.get_data_for_mod_entries_endpoint(courses),
        }

        events = (await self.client.async_post('core_calendar_get_calendar_events', calendar_req_data)).get(
            'events', []
        )

        events_per_course = self.sort_by_courseid(events)

        for course_id, events in events_per_course.items():
            event_files = []
            for event in events:
                event_name = event.get('name', 'unnamed event')
                event_description = event.get('description', None)

                event_modulename = event.get('modulename', None)
                event_timestart = event.get('timestart', 0)
                event_timeduration = event.get('timeduration', 0)

                event_filename = PT.to_valid_name(
                    f'{datetime.fromtimestamp(event_timestart).strftime("%Y.%m.%d %H:%M")} {event_name}', is_file=False
                )
                event_content = moodle_event_header
                event_content += f'<div class="event-title"><span class="icon">&#128197;</span>{event_name}</div>'
                event_content += (
                    '<div class="attribute"><span class="icon">&#9201;</span>'
                    + f'Start Time: {datetime.fromtimestamp(event_timestart).strftime("%c")}</div>'
                )
                if event_timeduration != 0:
                    event_timeend = event_timestart + event_timeduration
                    event_content += (
                        '<div class="attribute"><span class="icon">&#9201;</span>'
                        + f'End Time: {datetime.fromtimestamp(event_timeend).strftime("%c")}</div>'
                    )
                if event_description is not None and event_description != '':
                    event_content += (
                        '<div class="attribute"><span class="icon">&#128196;</span>' + f'{event_description}</div>'
                    )
                if event_modulename is not None:
                    event_content += (
                        '<div class="attribute"><span class="icon">&#128218;</span>'
                        f'Module Type: {event_modulename}</div>'
                    )

                event_content += moodle_event_footer

                event_files.append(
                    {
                        'filename': event_filename,
                        'filepath': '/',
                        'html': event_content,
                        'type': 'html',
                        'timemodified': event.get('timemodified', 0),
                        'filesize': len(event_content),
                        'no_search_for_urls': True,
                    }
                )
            if course_id not in core_contents:
                core_contents[course_id] = {}
            core_contents[course_id].append(
                {
                    'id': course_events_section_id,
                    'name': 'Events',
                    'modules': [{'id': course_events_module_id, 'name': 'Events', 'modname': 'calendar'}],
                }
            )

            self.add_module(
                result,
                course_id,
                course_events_module_id,
                {
                    'id': course_events_module_id,
                    'name': 'Events',
                    'files': event_files,
                },
            )

        return result

    @staticmethod
    def sort_by_courseid(events):
        sorted_dict = {}
        for event in events:
            course_id = event.get('courseid', 0)
            if course_id not in sorted_dict:
                sorted_dict[course_id] = []
            sorted_dict[course_id].append(event)
        return sorted_dict



import logging
import traceback
from typing import List

from moodle_dl.downloader.task import Task
from moodle_dl.notifications.notification_service import NotificationService
from moodle_dl.notifications.xmpp.xmpp_formater import XmppFormater as XF
from moodle_dl.notifications.xmpp.xmpp_shooter import XmppShooter
from moodle_dl.types import Course


class XmppService(NotificationService):
    def _is_configured(self) -> bool:
        # Checks if the sending of XMPP messages has been configured.
        try:
            self.config.get_property('xmpp')
            return True
        except ValueError:
            logging.debug('XMPP-Notifications not configured, skipping.')
            return False

    def _send_messages(self, messages: List[str]):
        """
        Sends an message
        """
        if not self._is_configured() or messages is None or len(messages) == 0:
            return

        xmpp_cfg = self.config.get_property('xmpp')

        logging.info('Sending Notification via XMPP...')

        xmpp_shooter = XmppShooter(xmpp_cfg['sender'], xmpp_cfg['password'], xmpp_cfg['target'])
        for message_content in messages:
            try:
                xmpp_shooter.send(message_content)
            except BaseException as e:
                logging.error('While sending notification:\n%s', traceback.format_exc(), extra={'exception': e})
                raise e  # to be properly notified via Sentry

    def notify_about_changes_in_moodle(self, changes: List[Course]) -> None:
        """
        Sends out a notification about the downloaded changes.
        @param changes: A list of changed courses with changed files.
        """
        if not self._is_configured():
            return

        messages = XF.create_full_moodle_diff_messages(changes)

        self._send_messages(messages)

    def notify_about_error(self, error_description: str):
        """
        Sends out an error message if configured to do so.
        @param error_description: The error object.
        """
        if not self._is_configured():
            return

        xmpp_cfg = self.config.get_property('xmpp')

        if not xmpp_cfg.get('send_error_msg', True):
            return
        messages = XF.create_full_error_messages(error_description)

        self._send_messages(messages)

    def notify_about_failed_downloads(self, failed_downloads: List[Task]):
        """
        Sends out an message about failed download if configured to send out error messages.
        @param failed_downloads: A list of failed Tasks.
        """
        if not self._is_configured():
            return

        xmpp_cfg = self.config.get_property('xmpp')

        if not xmpp_cfg.get('send_error_msg', True):
            return
        messages = XF.create_full_failed_downloads_messages(failed_downloads)

        self._send_messages(messages)



import threading
from dataclasses import dataclass, field
from enum import Enum
from typing import Any, Dict, List

from moodle_dl.utils import PathTools as PT


class File:
    def __init__(
        self,
        module_id: int,
        section_name: str,
        section_id: int,
        module_name: str,
        content_filepath: str,
        content_filename: str,
        content_fileurl: str,
        content_filesize: int,
        content_timemodified: int,
        module_modname: str,
        content_type: str,
        content_isexternalfile: bool,
        saved_to: str = '',
        time_stamp: int = 0,
        modified: int = 0,
        moved: int = 0,
        deleted: int = 0,
        notified: int = 0,
        file_hash: str = None,
        file_id: int = None,
        old_file_id: int = None,
    ):
        self.file_id = file_id

        self.module_id = module_id
        self.section_name = section_name
        self.section_id = section_id
        self.module_name = module_name

        self.content_filepath = content_filepath
        self.content_filename = content_filename
        self.content_fileurl = content_fileurl
        self.content_filesize = content_filesize
        self.content_timemodified = 0
        if content_timemodified is not None:
            self.content_timemodified = int(content_timemodified)

        self.module_modname = module_modname
        self.content_type = content_type

        if isinstance(content_isexternalfile, bool):
            self.content_isexternalfile = content_isexternalfile
        else:
            if content_isexternalfile == 1:
                self.content_isexternalfile = True
            else:
                self.content_isexternalfile = False

        self.saved_to = saved_to

        self.time_stamp = time_stamp

        if modified == 1:
            self.modified = True
        else:
            self.modified = False

        if moved == 1:
            self.moved = True
        else:
            self.moved = False

        if deleted == 1:
            self.deleted = True
        else:
            self.deleted = False

        if notified == 1:
            self.notified = True
        else:
            self.notified = False

        self.hash = file_hash

        # For text label
        self.text_content = None

        # For Created HTML-Files like Quizzes
        self.html_content = None

        # To manage the corresponding moved or changed files
        self.old_file = None
        self.new_file = None

        self.old_file_id = old_file_id

    def getMap(self) -> {str: str}:
        return {
            'file_id': self.file_id,
            'module_id': self.module_id,
            'section_name': self.section_name,
            'section_id': self.section_id,
            'module_name': self.module_name,
            'content_filepath': self.content_filepath,
            'content_filename': self.content_filename,
            'content_fileurl': self.content_fileurl,
            'content_filesize': self.content_filesize,
            'content_timemodified': self.content_timemodified,
            'module_modname': self.module_modname,
            'content_type': self.content_type,
            'content_isexternalfile': 1 if self.content_isexternalfile else 0,
            'saved_to': self.saved_to,
            'time_stamp': self.time_stamp,
            'modified': 1 if self.modified else 0,
            'moved': 1 if self.moved else 0,
            'deleted': 1 if self.deleted else 0,
            'notified': 1 if self.notified else 0,
            'hash': self.hash,
            'old_file_id': self.old_file_id,
        }

    @staticmethod
    def fromRow(row):
        return File(
            file_id=row['file_id'],
            module_id=row['module_id'],
            section_name=row['section_name'],
            section_id=row['section_id'],
            module_name=row['module_name'],
            content_filepath=row['content_filepath'],
            content_filename=row['content_filename'],
            content_fileurl=row['content_fileurl'],
            content_filesize=row['content_filesize'],
            content_timemodified=row['content_timemodified'],
            module_modname=row['module_modname'],
            content_type=row['content_type'],
            content_isexternalfile=row['content_isexternalfile'],
            saved_to=row['saved_to'],
            time_stamp=row['time_stamp'],
            modified=row['modified'],
            moved=row['moved'],
            deleted=row['deleted'],
            notified=row['notified'],
            file_hash=row['hash'],
            old_file_id=row['old_file_id'],
        )

    INSERT = """INSERT INTO files
            (course_id, course_fullname, module_id, section_name, section_id,
            module_name, content_filepath, content_filename,
            content_fileurl, content_filesize, content_timemodified,
            module_modname, content_type, content_isexternalfile,
            saved_to, time_stamp, modified, moved, deleted, notified,
            hash, old_file_id)
            VALUES (:course_id, :course_fullname, :module_id,
            :section_name, :section_id, :module_name, :content_filepath,
            :content_filename, :content_fileurl, :content_filesize,
            :content_timemodified, :module_modname, :content_type,
            :content_isexternalfile, :saved_to, :time_stamp,
            :modified, :moved, :deleted, :notified,  :hash,
            :old_file_id);
            """

    def __str__(self):
        message = 'File ('

        message += f'module_id: {self.module_id}'
        message += f', section_name: "{PT.to_valid_name(self.section_name, is_file=False)}"'
        message += f', section_id: "{self.section_id}"'
        message += f', module_name: "{PT.to_valid_name(self.module_name, is_file=False)}"'
        message += f', content_filepath: {self.content_filepath}'
        valid_content_filename = PT.to_valid_name(self.content_filename, is_file=True)
        if len(valid_content_filename) > 256:
            message += (
                f', content_filename (longer then 256 chars): "{valid_content_filename[:200]}[...]'
                + f'{valid_content_filename[-50:]}"'
            )
        else:
            message += f', content_filename: "{valid_content_filename}"'
        if len(self.content_fileurl) > 256:
            message += (
                f', content_fileurl (longer then 256 chars): "{self.content_fileurl[:200]}[...]'
                + f'{self.content_fileurl[-50:]}"'
            )
        else:
            message += f', content_fileurl: "{self.content_fileurl}"'
        message += f', content_filesize: {self.content_filesize}'
        message += f', content_timemodified: {self.content_timemodified}'
        message += f', module_modname: {self.module_modname}'
        message += f', content_type: {self.content_type}'
        message += f', content_isexternalfile: {self.content_isexternalfile}'

        message += f', saved_to: "{self.saved_to}"'
        message += f', time_stamp: {self.time_stamp}'
        message += f', modified: {self.modified}'
        message += f', moved: {self.moved}'
        message += f', deleted: {self.deleted}'
        message += f', notified: {self.notified}'
        message += f', hash: {self.hash}'
        message += f', file_id: {self.file_id}'
        message += f', old_file_id: {self.old_file_id}'

        message += ')'
        return message


class Course:
    def __init__(self, _id: int, fullname: str, files: List[File] = None):
        self.id = _id
        self.fullname = PT.to_valid_name(fullname, is_file=False)
        if files is not None:
            self.files = files
        else:
            self.files = []

        self.overwrite_name_with = None
        self.create_directory_structure = True
        self.excluded_sections = []

    def __str__(self):
        message = 'Course ('

        message += f'id: {self.id}'
        message += f', fullname: "{self.fullname}"'
        message += f', overwrite_name_with: "{PT.to_valid_name(self.overwrite_name_with, is_file=False)}"'
        message += f', create_directory_structure: {self.create_directory_structure}'
        message += f', files: {len(self.files)}'
        message += ')'
        return message


@dataclass
class MoodleURL:
    use_http: bool
    domain: str
    path: str
    scheme: str = field(init=False)
    url_base: str = field(init=False)

    def __post_init__(self):
        if self.use_http:
            self.scheme = 'http://'
        else:
            self.scheme = 'https://'
        self.url_base = self.scheme + self.domain + self.path


@dataclass
class MoodleDlOpts:
    init: bool
    config: bool
    new_token: bool
    change_notification_mail: bool
    change_notification_telegram: bool
    change_notification_discord: bool
    change_notification_xmpp: bool
    manage_database: bool
    delete_old_files: bool
    log_responses: bool
    add_all_visible_courses: bool
    sso: bool
    username: str
    password: str
    token: str
    path: str
    max_parallel_api_calls: int
    max_parallel_downloads: int
    max_parallel_yt_dlp: int
    download_chunk_size: int
    ignore_ytdl_errors: bool
    without_downloading_files: bool
    max_path_length_workaround: bool
    allow_insecure_ssl: bool
    skip_cert_verify: bool
    verbose: bool
    quiet: bool
    log_to_file: bool
    log_file_path: str


class TaskState(Enum):
    INIT = 'INIT'
    STARTED = 'STARTED'
    FAILED = 'FAILED'
    FINISHED = 'FINISHED'


@dataclass
class TaskStatus:
    state: TaskState = field(init=False, default=TaskState.INIT)
    bytes_downloaded: int = field(init=False, default=0)
    external_total_size: int = field(init=False, default=0)
    error: Any = field(init=False, default=None)
    yt_dlp_failed_with_error: bool = field(init=False, default=False)
    yt_dlp_used_generic_extractor: bool = field(init=False, default=False)
    yt_dlp_current_file: str = field(init=False, default=None)
    yt_dlp_total_size_per_file: Dict[str, int] = field(init=False, default_factory=dict)
    yt_dlp_bytes_downloaded_per_file: Dict[str, int] = field(init=False, default_factory=dict)

    def get_error_text(self) -> str:
        str_error = str(self.error).strip()
        if str_error != '':
            return str_error
        return repr(self.error)


@dataclass
class DownloadStatus:
    bytes_downloaded: int = field(init=False, default=0)
    bytes_to_download: int = field(init=False, default=0)

    files_downloaded: int = field(init=False, default=0)
    files_failed: int = field(init=False, default=0)
    files_to_download: int = field(init=False, default=0)

    lock: threading.Lock = field(init=False, default_factory=threading.Lock)


class DlEvent(Enum):
    FINISHED = 'FINISHED'
    FAILED = 'FAILED'
    RECEIVED = 'RECEIVED'
    TOTAL_SIZE = 'TOTAL_SIZE'
    TOTAL_SIZE_UPDATE = 'TOTAL_SIZE_UPDATE'


@dataclass
class DownloadOptions:
    token: str
    download_linked_files: bool
    download_domains_whitelist: List
    download_domains_blacklist: List
    cookies_text: str
    yt_dlp_options: Dict
    video_passwords: Dict
    external_file_downloaders: Dict
    restricted_filenames: bool
    write_links: Dict
    download_path: str
    global_opts: MoodleDlOpts


@dataclass
class HeadInfo:
    content_type: str
    is_html: bool = field(init=False, default=False)
    content_length: int
    last_modified: str
    final_url: str
    guessed_file_name: str
    host: str

    def __post_init__(self):
        if self.content_type in ('text/html', 'text/plain'):
            self.is_html = True
